<?php

namespace Acms\Plugins\ShoppingCart\GET\Payment;

use Acms\Plugins\ShoppingCart\Entities\Master\Status;
use Acms\Plugins\ShoppingCart\Entities\Order;
use Acms\Plugins\ShoppingCart\Entities\Payment;
use Acms\Plugins\ShoppingCart\Repositories\ConfigRepository;
use Acms\Plugins\ShoppingCart\Services\ShoppingService;
use App;
use ACMS_GET;
use ACMS_Corrector;
use Template;
use Acms\Services\Facades\Media;

/**
 * 決済フォームで選択可能な決済方法を表示するモジュール
 */
class Select extends ACMS_GET
{
    protected const MEDIA_FIELD_PREFIX = 'image';

    /**
     * @var ShoppingService
     */
    protected ShoppingService $ShoppingService;

    /**
     * @var ConfigRepository
     */
    protected ConfigRepository $Config;

    /**
     * @var Order|null
     */
    protected $Order;

    public function get()
    {
        $Tpl = new Template($this->tpl, new ACMS_Corrector());
        $this->buildModuleField($Tpl);

        $this->ShoppingService = App::make('shopping-cart.service.shopping');
        $this->Config = App::make('shopping-cart.repository.config');

        /**
         * Set Order
         */
        $this->setOrder();

        if (is_null($this->Order)) {
            return $Tpl->render([
                'notFound' => new \stdClass()
            ]);
        }

        $Payments = $this->getPayments();

        if (empty($Payments)) {
            return $Tpl->render([
                'notFound' => new \stdClass()
            ]);
        }

        return $Tpl->render([
            'payment' => $this->buildPayments($Payments)
        ]);
    }

    /**
     * 購入処理中の注文情報を取得する
     *
     * @return Order|null
     */
    public function loadProcessingOrder(): ?Order
    {
        $processingOrderId = $this->Post->get('processingOrderId');

        if ($processingOrderId !== '') {
            return $this->ShoppingService->loadProcessingOrder($processingOrderId);
        }

        return null;
    }


    /**
     * 決済フォームで利用可能な決済方法を取得する
     *
     * @return Payment[]
     */
    protected function getPayments(): array
    {
        return $this->ShoppingService->getAvailablePayments(
            $this->Config->findPayments(null, Status::OPEN),
            $this->Order
        );
    }

    /**
     * 決済方法のテンプレートを組み立て
     *
     * @param Payment[] $Payments
     *
     * @return array
     */
    protected function buildPayments($Payments): array
    {
        return array_map(
            function (Payment $Payment) {
                return array_merge(
                    [
                        'block' => $Payment->getBlock(),
                        'status' => $Payment->getStatus(),
                        'name' => $Payment->getName(),
                        'commissionPaid' => $Payment->getCommissionPaid(),
                        'isRuleEnabled' => $Payment->getIsRuleEnabled(),
                        'description' => $Payment->getDescription(),
                        'mailMessage' => $Payment->getMailMessage(),
                    ],
                    (!is_null($Payment->getMedia()) ? [
                        $this::MEDIA_FIELD_PREFIX => $Payment->getMedia()->getId(),
                        $this::MEDIA_FIELD_PREFIX . '@name' => $Payment->getMedia()->getFileName(),
                        $this::MEDIA_FIELD_PREFIX . '@extension' => $Payment->getMedia()->getExtension(),
                        $this::MEDIA_FIELD_PREFIX . '@fileSize' => $Payment->getMedia()->getFileSize(),
                        $this::MEDIA_FIELD_PREFIX . '@caption' => $Payment->getMedia()->getCaption(),
                        $this::MEDIA_FIELD_PREFIX . '@link' => $Payment->getMedia()->getLink(),
                        $this::MEDIA_FIELD_PREFIX . '@alt' => $Payment->getMedia()->getAlt(),
                        $this::MEDIA_FIELD_PREFIX . '@text' => $Payment->getMedia()->getText(),
                        $this::MEDIA_FIELD_PREFIX . '@path' => $Payment->getMedia()->getPath(),
                        $this::MEDIA_FIELD_PREFIX . '@thumbnail' => Media::getImageThumbnail(
                            $Payment->getMedia()->getPath()
                        ),
                        $this::MEDIA_FIELD_PREFIX . '@imageSize' => $Payment->getMedia()->getImageSize(),
                        $this::MEDIA_FIELD_PREFIX . '@type' => $Payment->getMedia()->getType(),
                        $this::MEDIA_FIELD_PREFIX . '@page' => $Payment->getMedia()->getPdfPage(),
                        $this::MEDIA_FIELD_PREFIX . '@focalX' => $Payment->getMedia()->getFocalX(),
                        $this::MEDIA_FIELD_PREFIX . '@focalY' => $Payment->getMedia()->getFocalY(),
                        $this::MEDIA_FIELD_PREFIX . '@width' => $Payment->getMedia()->getWidth(),
                        $this::MEDIA_FIELD_PREFIX . '@height' => $Payment->getMedia()->getHeight(),
                        $this::MEDIA_FIELD_PREFIX . '@ratio' => $Payment->getMedia()->getRatio(),
                    ] : [])
                );
            },
            $Payments
        );
    }

    /**
     * 注文情報のセット
     *
     * @return void
     */
    protected function setOrder()
    {
        $this->Order = $this->loadProcessingOrder();
    }
}
