<?php

namespace Acms\Plugins\ShoppingCart\GET\ShoppingCart;

use Acms\Plugins\ShoppingCart\Entities\Addition;
use Acms\Plugins\ShoppingCart\Entities\Payment;
use Acms\Plugins\ShoppingCart\Entities\PaymentRule;
use Acms\Plugins\ShoppingCart\Entities\Shipping;
use Acms\Plugins\ShoppingCart\Entities\ShippingGroup;
use Acms\Plugins\ShoppingCart\Repositories\ConfigRepository;
use App;
use ACMS_GET;
use ACMS_Corrector;
use Template;
use Media;

/**
 * ShoppingCart拡張アプリの設定を表示する
 */
class Settings extends ACMS_GET
{
    protected const MEDIA_FIELD_PREFIX = 'image';

    /**
     * @var ConfigRepository
     */
    protected $Config;

    public function get()
    {
        $Tpl = new Template($this->tpl, new ACMS_Corrector());
        $this->buildModuleField($Tpl);

        $this->Config = App::make('shopping-cart.repository.config');

        return $Tpl->render([
            'tax' => $this->buildTax(),
            'payment' => $this->buildPayment(),
            'shipping' => $this->buildShipping(),
            'addition' => $this->buildAddition()
        ]);
    }

    /**
     * 税金ブロックの組み立て
     *
     * @return array
     */
    protected function buildTax(): array
    {
        return [
            'expressionFormat' => $this->Config->get('shopping-cart_tax-format', ''),
            'calcMethod' => $this->Config->get('shopping-cart_tax-calc-method', ''),
            'roundingMethod' => $this->Config->get('shopping-cart_tax-rounding', ''),
            'invoiceNumber' => $this->Config->get('shopping_cart_invoice-number', ''),
            'invoiceName' => $this->Config->get('shopping_cart_invoice-name', '')
        ];
    }

    /**
     * 決済方法ブロックの組み立て
     *
     * @return array
     */
    protected function buildPayment(): array
    {
        return array_map(
            function (Payment $Payment) {
                return array_merge(
                    [
                        'block' => $Payment->getBlock(),
                        'status' => $Payment->getStatus(),
                        'name' => $Payment->getName(),
                        'commissionPaid' => $Payment->getCommissionPaid(),
                        'isRuleEnabled' => $Payment->getIsRuleEnabled(),
                        'description' => $Payment->getDescription(),
                        'mailMessage' => $Payment->getMailMessage(),
                    ],
                    (!is_null($Payment->getMedia()) ? [
                        $this::MEDIA_FIELD_PREFIX => $Payment->getMedia()->getId(),
                        $this::MEDIA_FIELD_PREFIX . '@name' => $Payment->getMedia()->getFileName(),
                        $this::MEDIA_FIELD_PREFIX . '@extension' => $Payment->getMedia()->getExtension(),
                        $this::MEDIA_FIELD_PREFIX . '@fileSize' => $Payment->getMedia()->getFileSize(),
                        $this::MEDIA_FIELD_PREFIX . '@caption' => $Payment->getMedia()->getCaption(),
                        $this::MEDIA_FIELD_PREFIX . '@link' => $Payment->getMedia()->getLink(),
                        $this::MEDIA_FIELD_PREFIX . '@alt' => $Payment->getMedia()->getAlt(),
                        $this::MEDIA_FIELD_PREFIX . '@text' => $Payment->getMedia()->getText(),
                        $this::MEDIA_FIELD_PREFIX . '@path' => $Payment->getMedia()->getPath(),
                        $this::MEDIA_FIELD_PREFIX . '@thumbnail' => Media::getImageThumbnail(
                            $Payment->getMedia()->getPath()
                        ),
                        $this::MEDIA_FIELD_PREFIX . '@imageSize' => $Payment->getMedia()->getImageSize(),
                        $this::MEDIA_FIELD_PREFIX . '@type' => $Payment->getMedia()->getType(),
                        $this::MEDIA_FIELD_PREFIX . '@page' => $Payment->getMedia()->getPdfPage(),
                        $this::MEDIA_FIELD_PREFIX . '@focalX' => $Payment->getMedia()->getFocalX(),
                        $this::MEDIA_FIELD_PREFIX . '@focalY' => $Payment->getMedia()->getFocalY(),
                        $this::MEDIA_FIELD_PREFIX . '@width' => $Payment->getMedia()->getWidth(),
                        $this::MEDIA_FIELD_PREFIX . '@height' => $Payment->getMedia()->getHeight(),
                        $this::MEDIA_FIELD_PREFIX . '@ratio' => $Payment->getMedia()->getRatio(),
                    ] : []),
                    ['paymentRule' => array_map(
                        function (PaymentRule $PaymentRule) use ($Payment) {
                            return [
                                'block' => $Payment->getBlock(),
                                'priceMin' => $PaymentRule->getPriceMin(),
                                'priceMax' => $PaymentRule->getPriceMax(),
                                'commission' => $PaymentRule->getCommission(),
                            ];
                        },
                        $Payment->getPaymentRules()
                    )
                    ]
                );
            },
            $this->Config->findAllPayments()
        );
    }

    /**
     * 配送料ブロックの組み立て
     *
     * @return array
     */
    protected function buildShipping()
    {
        return [
            'dateFrom' => $this->Config->get('shopping-cart_shipping-date-from'),
            'dateTo' => $this->Config->get('shopping-cart_shipping-date-to'),
            'timeframe' => $this->Config->getArray('shopping-cart_shipping-time-frame'),
            'shippingFreeSubtotal' => $this->Config->get('shopping-cart_shipping-free-subtotal'),
            'calcMethod' => $this->Config->get('shopping-cart_calc-method', ''),
            'group' => $this->buildShippingGroup()
        ];
    }

    /**
     * 配送グループブロックの組み立て
     *
     * @return array
     */
    protected function buildShippingGroup()
    {
        return array_map(
            function (ShippingGroup $ShippingGroup) {
                return [
                    'id' => $ShippingGroup->getId(),
                    'status' => $ShippingGroup->getStatus(),
                    'name' => $ShippingGroup->getName(),
                    'groupFee' => $ShippingGroup->getGroupFee(),
                    'calcMethod' => $ShippingGroup->getCalcMethod(),
                    'region' => array_map(
                        function (Shipping $Shipping) use ($ShippingGroup) {
                            return [
                                'gid' => $ShippingGroup->getId(),
                                'regionName' => $Shipping->getName(),
                                'regionFee' => $Shipping->getFee()
                            ];
                        },
                        $ShippingGroup->getShippings()
                    )
                ];
            },
            $this->Config->findAllShippingGroups()
        );
    }

    /**
     * 追加手数料ブロックの組み立て
     *
     * @return array
     */
    protected function buildAddition()
    {
        return array_map(
            function (Addition $Addition) {
                return [
                    'id' => $Addition->getId(),
                    'status' => $Addition->getStatus(),
                    'name' => $Addition->getName(),
                    'fee' => $Addition->getFee()
                ];
            },
            $this->Config->findAllAdditions()
        );
    }
}
