<?php

namespace Acms\Plugins\ShoppingCart\GET\ShoppingCart;

use Acms\Plugins\ShoppingCart\Entities\Cart;
use Acms\Plugins\ShoppingCart\Entities\Item;
use Acms\Plugins\ShoppingCart\Entities\TaxAmount;
use Acms\Plugins\ShoppingCart\Entities\ShippingGroup;
use Acms\Plugins\ShoppingCart\Repositories\ConfigRepository;
use Acms\Plugins\ShoppingCart\Repositories\CartRepository;
use Acms\Plugins\ShoppingCart\Services\TaxService;
use ACMS_GET_Entry_Summary;
use App;
use Acms\Services\Facades\Database as DB;
use Template;

/**
 * カートに入っているアイテム一覧を表示する
 */
class Summary extends ACMS_GET_Entry_Summary
{
    /**
     * @return Cart
     */
    protected Cart $Cart;

    /**
     * @return TaxAmount[]
     */
    protected array $TaxAmounts = [];

    /**
     * @var CartRepository
     */
    protected CartRepository $CartRepository;

    /**
     * @var TaxService
     */
    protected TaxService $TaxService;

    /**
     * @var ConfigRepository
     */
    protected ConfigRepository $Config;

    /**
     * 数量をエントリーのデータに追加する際に使用するキー
     *
     * @var string
     */
    protected const QUANTITY_KEY = 'item_quantity';

    /**
     * 商品価格をエントリーのデータに追加する際に使用するキー
     *
     * @var string
     */
    protected const PRICE_KEY = 'item_price';

    /**
     * 税率をエントリーのデータに追加する際に使用するキー
     *
     * @var string
     */
    protected const TAX_RATE_KEY = 'item_tax_rate';

    /**
     * 在庫数をエントリーのデータに追加する際に使用するキー
     *
     * @var string
     */
    protected const STOCK_KEY = 'item_stock';

    /**
     * 配送グループをエントリーのデータに追加する際に使用するキー
     *
     * @var string
     */
    protected const SHIPPING_GROUP_KEY = 'item_shipping_group';

    /**
     * 税込合計金額をエントリーのデータに追加する際に使用するキー
     *
     * @var string
     */
    protected const TOTAL_TAX_INCLUDED_KEY = 'item_total_tax_included';

    /**
     * 消費税額をエントリーのデータに追加する際に使用するキー
     *
     * @var string
     */
    protected const TAX_PRICE_KEY = 'item_tax_price';

    /**
     * コンフィグの取得
     *
     * @return array
     */
    public function initVars()
    {
        return [
            'order' => [
                $this->order ? $this->order : config('shoppingcart_summary_order'),
                config('shoppingcart_summary_order2'),
            ],
            'orderFieldName'        => config('shoppingcart_summary_order_field_name'),
            'limit'                 => PHP_INT_MAX,
            'indexing'              => config('shoppingcart_summary_indexing'),
            'subCategory'           => config('shoppingcart_summary_sub_category'),
            'secret'                => config('shoppingcart_summary_secret'),
            'notfound'              => config('mo_shoppingcart_summary_notfound'),
            'notfoundStatus404'     => config('shoppingcart_summary_notfound_status_404'),
            'noimage'               => config('shoppingcart_summary_noimage'),

            'unit'                  => config('shoppingcart_summary_unit'),
            'newtime'               => config('shoppingcart_summary_newtime'),
            'imageX'                => intval(config('shoppingcart_summary_image_x')),
            'imageY'                => intval(config('shoppingcart_summary_image_y')),
            'imageTrim'             => config('shoppingcart_summary_image_trim'),
            'imageZoom'             => config('shoppingcart_summary_image_zoom'),
            'imageCenter'           => config('shoppingcart_summary_image_center'),

            'entryFieldOn'          => config('shoppingcart_summary_entry_field'),
            'relatedEntryOn'        => config('shoppingcart_summary_related_entry_on', 'off'),
            'categoryInfoOn'        => config('shoppingcart_summary_category_on'),
            'categoryFieldOn'       => config('shoppingcart_summary_category_field_on'),
            'userInfoOn'            => config('shoppingcart_summary_user_on'),
            'userFieldOn'           => config('shoppingcart_summary_user_field_on'),
            'blogInfoOn'            => config('shoppingcart_summary_blog_on'),
            'blogFieldOn'           => config('shoppingcart_summary_blog_field_on'),
            'mainImageOn'           => config('shoppingcart_summary_image_on'),
            'detailDateOn'          => config('shoppingcart_summary_date'),
            'fullTextOn'            => config('shoppingcart_summary_fulltext'),
            'fulltextWidth'         => config('shoppingcart_summary_fulltext_width'),
            'fulltextMarker'        => config('shoppingcart_summary_fulltext_marker'),
            'tagOn'                 => config('shoppingcart_summary_tag'),
            'hiddenPrivateEntry'    => config('shoppingcart_summary_hidden_private_entry'),
            'loop_class'            => config('shoppingcart_summary_loop_class'),
            'relational'            => 'on',
        ];
    }

    public function get()
    {
        if (!defined('NO_CACHE_PAGE')) {
            define('NO_CACHE_PAGE', 1);
        }
        $this->TaxService = App::make('shopping-cart.service.tax');
        $this->CartRepository = App::make('shopping-cart.repository.cart');
        $this->Config = App::make('shopping-cart.repository.config');
        return parent::get();
    }

    // To-Do: カートへの追加順（昇順）以外での表示順対応
    // To-Do: 絞り込み機能（期間、非公開記事は現状でも絞り込み可能）

    /**
     * limitクエリ組み立て
     *
     * @param \SQL_Select & $SQL
     * @return void
     */
    public function limitQuery(&$SQL)
    {
        return true; // limitはしない
    }

    /**
     * カート内エントリーの取得
     *
     * @return bool
     */
    public function setRelational()
    {
        $this->Cart = $this->CartRepository->find($this->bid);
        $CartItems = $this->Cart->getItems();
        if (empty($CartItems)) {
            $this->eids = [];
            return true;
        }

        $this->eids = array_map(
            function (Item $item) {
                return $item->getEntry()->getId();
            },
            $CartItems
        );

        $this->TaxAmounts = $this->TaxService->createTaxAmounts($CartItems);

        return true;
    }

    /**
     * テンプレートの組み立て
     *
     * @param Template & $Tpl
     * @return array
     */
    public function buildEntries(&$Tpl)
    {
        $gluePoint = count($this->entries);
        $eagerLoad = $this->eagerLoad();
        $this->buildTaxAmount($Tpl);

        foreach ($this->entries as $i => $row) {
            $i++;

            $cartItem = $this->Cart->getItemByEntryId(intval($row['entry_id']));

            if (!is_null($cartItem)) {
                $Tpl->add(
                    ['cartField', 'entry:loop'],
                    $this->buildField($cartItem->getField(), $Tpl, ['entry:loop'])
                );
                if (!is_null($cartItem->getEntry()->getShippingGroup())) {
                    $this->buildShippingGroup($Tpl, $cartItem->getEntry()->getShippingGroup());
                }
                $row += $this->getCartItemVars($cartItem);
            }

            $this->buildSummary($Tpl, $row, $i, $gluePoint, $this->config, $this->getExtraVars(), $eagerLoad);
        }
    }

    /**
     * ルート変数の取得
     *
     * @return array
     */
    public function getRootVars()
    {
        return array_merge(
            parent::getRootVars(),
            [
                'itemsAmount' => intval(DB::query($this->amount->get($this->dsn()), 'one')),
                'quantitiesAmount' => array_reduce(
                    $this->Cart->getItems(),
                    function (int $quantity, Item $Item) {
                        return $quantity + $Item->getQuantity();
                    },
                    0
                ),
                'subtotal' => array_reduce(
                    $this->TaxAmounts,
                    function ($subtotal, TaxAmount $TaxAmount) {
                        return $subtotal + $TaxAmount->getExcludingTax() + $TaxAmount->getTax();
                    },
                    0
                )
            ]
        );
    }

    /**
     * エントリーフィールドのEagerLoading
     *
     * @return array|bool
     */
    protected function entryFieldEagerLoad()
    {
        if (!isset($this->config['entryFieldOn']) || $this->config['entryFieldOn'] === 'on') {
            return array_reduce(
                $this->Cart->getItems(),
                function (array $array, Item $Item) {
                    return $array + [$Item->getEntry()->getId() => $Item->getEntry()->getField()];
                },
                []
            );
        }
        return false;
    }

    /**
     * カートアイテム変数の取得
     *
     * @param Item $item
     *
     * @return array
     */
    protected function getCartItemVars(Item $item)
    {
        $price = intval($item->getEntry()->getField()->get('item_price'));
        $quantity = $item->getQuantity();
        $taxRate = floatval($item->getEntry()->getField()->get('item_tax'));
        $stock = $item->isStockManagementEnabled()
            ? intval($item->getEntry()->getField()->get('item_stock'))
            : null;

        $isExTax = $this->Config->get('shopping-cart_tax-format') === 'extax';
        $totalTaxIncluded = $this->TaxService->calcTotalPrice(
            $price,
            $quantity,
            $isExTax ? $taxRate : null
        );
        $taxPrice = $this->TaxService->calcTax($price, $taxRate);

        return [
            self::QUANTITY_KEY => $quantity,
            self::PRICE_KEY => $price,
            self::TAX_RATE_KEY => $taxRate,
            self::STOCK_KEY => $stock,
            self::TOTAL_TAX_INCLUDED_KEY => $totalTaxIncluded,
            self::TAX_PRICE_KEY => $taxPrice
        ];
    }

    /**
     * 追加変数の取得
     *
     * @return array
     */
    protected function getExtraVars()
    {
        return [
            'quantity' => self::QUANTITY_KEY,
            'price' => self::PRICE_KEY,
            'taxRate' => self::TAX_RATE_KEY,
            'stock' => self::STOCK_KEY,
            'totalTaxIncluded' => self::TOTAL_TAX_INCLUDED_KEY,
            'taxPrice' => self::TAX_PRICE_KEY
        ];
    }

    /**
     * 税金ブロックの組み立て
     *
     * @param Template $Tpl
     *
     * @return void
     */
    protected function buildTaxAmount(Template $Tpl): void
    {
        foreach ($this->TaxAmounts as $TaxAmount) {
            /** @var TaxAmount $TaxAmount */
            $Tpl->add(
                ['taxAmount:loop'],
                [
                    'taxRate' => $TaxAmount->getTaxRate() * 100,  // %に変換
                    'excludingTax' => $TaxAmount->getExcludingTax(),
                    'tax' => $TaxAmount->getTax()
                ]
            );
        }
    }

    /**
     * 配送グループブロックの組み立て
     *
     * @param Template $Tpl
     * @param ShippingGroup $ShippingGroup
     *
     * @return void
     */
    protected function buildShippingGroup(Template $Tpl, ShippingGroup $ShippingGroup): void
    {
        $Tpl->add(
            ['shippingGroup', 'entry:loop'],
            [
                'id' => $ShippingGroup->getId(),
                'status' => $ShippingGroup->getStatus(),
                'name' => $ShippingGroup->getName(),
                'groupFee' => $ShippingGroup->getGroupFee(),
                'calcMethod' => $ShippingGroup->getCalcMethod(),
            ]
        );
    }
}
