<?php

namespace Acms\Plugins\ShoppingCart\GET;

use Acms\Plugins\ShoppingCart\Entities\Cart;
use Acms\Plugins\ShoppingCart\Entities\Item;
use Acms\Plugins\ShoppingCart\Entities\Order;
use Acms\Plugins\ShoppingCart\Entities\Addition;
use Acms\Plugins\ShoppingCart\Repositories\CartRepository;
use Acms\Plugins\ShoppingCart\Services\ShoppingService;
use ACMS_Corrector;
use ACMS_GET_Form;
use App;
use Template;
use Field;
use Acms\Services\Facades\Media;

class ShoppingForm extends ACMS_GET_Form
{
    protected const MEDIA_FIELD_PREFIX = 'image';

    /**
     * @var CartRepository
     */
    protected $CartRepository;

    /**
     * @var ShoppingService
     */
    protected $ShoppingService;

    /**
     * @var Order|null
     */
    protected $Order = null;

    /**
     * @var Cart|null
     */
    protected $Cart = null;

    /**
     * @var string
     */
    public $step;

    /**
     * 初期化処理
     */
    protected function init()
    {
        $this->CartRepository = App::make('shopping-cart.repository.cart');
        $this->ShoppingService = App::make('shopping-cart.service.shopping');

        $this->Cart = $this->CartRepository->find(BID);
    }

    public function get()
    {
        if (!defined('NO_CACHE_PAGE')) {
            define('NO_CACHE_PAGE', 1);
        }
        $this->init();
        $Tpl = new Template($this->tpl, new ACMS_Corrector());

        $this->step = $this->Post->get('error');
        if (empty($this->step)) {
            $this->step = $this->Get->get('step');
        }
        if ($this->Post->isValidAll()) {
            $this->step = $this->Post->get('step', $this->step);
        } else {
            $this->error($Tpl);
        }
        if (!ACMS_POST) {
            $this->step = 'step';
        }
        $this->Post->delete('step');

        // 初期ステップかつ、
        // プロセス中の注文が存在しない場合（初期ステップの場合でもモジュールが2回目以降の実行の場合はすでにプロセス中の注文が存在する）は、
        // 注文情報の初期化処理
        if ($this->step === 'step' && is_null($this->getOrderProcessingId())) {
            $this->Order = $this->createOrder();
            $this->saveProcessingOrder($this->Order);
        }

        if (is_null($this->Order)) {
            $this->Order = $this->loadProcessingOrder();
        }

        $block = $this->getBlock();

        if (is_null($this->Order)) {
            return $Tpl->render([
                $block => $this->buildField($this->Post, $Tpl, $block, '')
            ]);
        }

        $this->mergeOrder($this->Order);

        return $Tpl->render([
            $block => array_merge(
                $this->buildTpl(),
                $this->buildField($this->Post, $Tpl, $block, '')
            )
        ]);
    }

    /**
     * 注文商品の取得
     *
     * @return Item[]
     */
    protected function getItems(): array
    {
        return $this->Cart->getItems();
    }

    /**
     * 受注処理中の注文情報IDの取得
     *
     * @return string|null
     */
    protected function getOrderProcessingId(): ?string
    {
        return $this->Cart->getOrderProcessingId();
    }

    /**
     * 購入処理中の注文情報を取得する
     *
     * @return Order|null
     */
    public function loadProcessingOrder(): ?Order
    {
        $orderProcessingId = $this->getOrderProcessingId();
        if (isset($orderProcessingId) && $orderProcessingId !== '') {
            return $this->ShoppingService->loadProcessingOrder($orderProcessingId);
        }

        return null;
    }

    /**
     * 購入処理中の注文情報を保存する
     *
     * @param Order $Order
     *
     * @return void
     */
    protected function saveProcessingOrder(Order $Order): void
    {
        $this->ShoppingService->saveProcessingOrder($Order);
        $this->saveOrderProcessingId($Order->getProcessingId());
    }

    /**
     * 受注処理中の注文情報IDの保存
     *
     * @param string $orderProcessingId
     * @return void
     */
    protected function saveOrderProcessingId(string $orderProcessingId): void
    {
        $this->Cart->setOrderProcessingId($orderProcessingId);
        $this->CartRepository->save($this->Cart, BID);
    }

    /**
     * 注文情報の作成
     *
     * @return Order
     */
    protected function createOrder(): Order
    {
        return $this->ShoppingService->createOrder($this->getItems());
    }

    /**
     * 表示するブロックを取得する
     *
     * @return string
     */
    protected function getBlock(): string
    {
        $block = !(empty($this->step) || is_bool($this->step)) ? 'step#' . $this->step : 'step';
        $pattern = '/<!--[\t 　]*BEGIN[\s]+' . preg_quote($block, '/') . '[\t 　]*-->/u';
        if (!multiBytePregMatch($pattern, $this->tpl)) {
            $block = 'step';
        }

        return $block;
    }

    /**
     * 注文情報のマージ
     * フォームを離脱した場合でも一度入力した注文データは保持する
     *
     * @param Order $Order
     *
     * @return void
     */
    protected function mergeOrder(Order $Order): void
    {
        $Field = new Field();
        if (!is_null($Order->getDeliveryArea())) {
            $Field->set('deliver-area', $Order->getDeliveryArea());
        }

        if (!is_null($Order->getPayment())) {
            $Field->set('payment-method', $Order->getPayment()->getBlock());
        }

        if (!empty($Order->getAdditions())) {
            $Field->set(
                'addition',
                array_map(
                    function (Addition $Addition) {
                        return $Addition->getId();
                    },
                    $Order->getAdditions()
                )
            );
        }

        $this->Post->overload($Field);
    }

    /**
     * テンプレートの組み立て
     *
     * @return array
     */
    protected function buildTpl(): array
    {
        return array_merge(
            (!is_null($this->Order->getPayment()) ? $this->buildPayment() : []),
            ['additionGroup' => (!empty($this->Order->getAdditions()) ? $this->buildAdditionGroup() : [])],
        );
    }

    /**
     * 適用中の決済方法のテンプレートを組み立て
     *
     * @return array
     */
    protected function buildPayment(): array
    {
        $MEIDIA_PREFIX = 'paymentMethod' . ucfirst($this::MEDIA_FIELD_PREFIX);

        return array_merge(
            [
                'paymentMethodBlock' => $this->Order->getPayment()->getBlock(),
                'paymentMethodName' => $this->Order->getPayment()->getName(),
                'paymentMethodCommissionPaid' => $this->Order->getPayment()->getCommissionPaid(),
                'paymentMethodDescription' => $this->Order->getPayment()->getDescription(),
                'paymentMethodMailMessage' => $this->Order->getPayment()->getMailMessage(),
            ],
            (!is_null($this->Order->getPayment()->getMedia()) ? [
                $MEIDIA_PREFIX => $this->Order->getPayment()->getMedia()->getId(),
                $MEIDIA_PREFIX . '@name' => $this->Order->getPayment()->getMedia()->getFileName(),
                $MEIDIA_PREFIX . '@extension' => $this->Order->getPayment()->getMedia()->getExtension(),
                $MEIDIA_PREFIX . '@fileSize' => $this->Order->getPayment()->getMedia()->getFileSize(),
                $MEIDIA_PREFIX . '@caption' => $this->Order->getPayment()->getMedia()->getCaption(),
                $MEIDIA_PREFIX . '@link' => $this->Order->getPayment()->getMedia()->getLink(),
                $MEIDIA_PREFIX . '@alt' => $this->Order->getPayment()->getMedia()->getAlt(),
                $MEIDIA_PREFIX . '@text' => $this->Order->getPayment()->getMedia()->getText(),
                $MEIDIA_PREFIX . '@path' => $this->Order->getPayment()->getMedia()->getPath(),
                $MEIDIA_PREFIX . '@thumbnail' => Media::getImageThumbnail(
                    $this->Order->getPayment()->getMedia()->getPath()
                ),
                $MEIDIA_PREFIX . '@imageSize' => $this->Order->getPayment()->getMedia()->getImageSize(),
                $MEIDIA_PREFIX . '@type' => $this->Order->getPayment()->getMedia()->getType(),
                $MEIDIA_PREFIX . '@page' => $this->Order->getPayment()->getMedia()->getPdfPage(),
                $MEIDIA_PREFIX . '@focalX' => $this->Order->getPayment()->getMedia()->getFocalX(),
                $MEIDIA_PREFIX . '@focalY' => $this->Order->getPayment()->getMedia()->getFocalY(),
                $MEIDIA_PREFIX . '@width' => $this->Order->getPayment()->getMedia()->getWidth(),
                $MEIDIA_PREFIX . '@height' => $this->Order->getPayment()->getMedia()->getHeight(),
                $MEIDIA_PREFIX . '@ratio' => $this->Order->getPayment()->getMedia()->getRatio(),
            ] : [])
        );
    }

    /**
     * 適用中の追加手数料のテンプレートを組み立て
     *
     * @return array
     */
    protected function buildAdditionGroup(): array
    {
        return array_map(
            function (Addition $Addition) {
                return [
                    'additionId' => $Addition->getId(),
                    'additionName' => $Addition->getName(),
                    'additionFee' => $Addition->getFee()
                ];
            },
            $this->Order->getAdditions()
        );
    }
}
