<?php

namespace Acms\Plugins\ShoppingCart\POST\ShoppingCart;

use Acms\Plugins\ShoppingCart\Entities\Item;
use Acms\Plugins\ShoppingCart\POST\ShoppingCart;
use ACMS_Validator;
use ACMS_POST_Cache;
use AcmsLogger;

/**
 * カートに商品を追加・更新する
 */
class AddItem extends ShoppingCart
{
    /**
     * @return Field_Validation|never
     */
    public function post()
    {
        $this->setBlock('add');

        $cart = $this->loadCart();

        $cartRequest = $this->extract('cart');
        $fieldRequest = $this->extract('field');

        $this->validateRequest($cartRequest, $fieldRequest);

        // 入力値のチェック
        if (!$this->Post->isValidAll()) {
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice('【Shopping Cart plugin】バリデートに失敗したため、カートへの商品追加に失敗しました', [
                    'cart' => $cartRequest->_aryV,
                    'field' => $fieldRequest->_aryV,
                ]);
            }
            return $this->Post;
        }

        $entryId = idval($cartRequest->get('eid'));
        $quantity = intval($cartRequest->get('quantity'));

        $item = $cart->getItemByEntryId($entryId);

        if (is_null($item)) {
            // 新規追加
            $entry = $this->EntryRepository->findById($entryId);

            if (is_null($entry)) {
                $cartRequest->setValidator('eid', 'exist', false);
                if (class_exists('AcmsLogger')) {
                    AcmsLogger::notice(
                        '【Shopping Cart plugin】指定したエントリーIDの商品が存在しないため、カートへの商品追加に失敗しました',
                        [
                            'entryId' => $entryId,
                        ]
                    );
                }
                return $this->Post;
            }

            $item = new Item();
            $item->setQuantity($quantity);
            $item->setEntry($entry);
            $item->setField($fieldRequest);

            $cart->addItem($item);
        } else {
            // 更新
            $item->addQuantity($quantity);
        }

        // 在庫チェック
        if ($item->isStockManagementEnabled() && $item->isStockShortage()) {
            $cartRequest->setValidator('quantity', 'stock', false);
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice(
                    '【Shopping Cart plugin】数量が在庫数を超過しているため、カートへの商品（' . $item->getEntry()->getTitle() . '）追加に失敗しました',
                    [
                        'entryId' => $entryId,
                        'quantity' => $quantity,
                        'stock' => intval($item->getEntry()->getField()->get('item_stock')),
                    ]
                );
            }
            return $this->Post;
        }

        $this->saveCart($cart);

        if (class_exists('AcmsLogger')) {
            AcmsLogger::info('【Shopping Cart plugin】カートへの商品（' . $item->getEntry()->getTitle() . '）追加に成功しました', [
                'entryId' => $entryId,
                'quantity' => $quantity,
            ]);
        }
        $this->setCartItem($item);

        if (config('cache_clear_when_post') === 'on') {
            ACMS_POST_Cache::clearPageCache(BID);
        }

        $this->redirect(HTTP_REQUEST_URL);
    }

    /**
     * 入力値のバリデート
     *
     * @param \Field_Validation $cartRequest
     * @param \Field_Validation $fieldRequest
     *
     * @return void
     */
    protected function validateRequest(\Field_Validation $cartRequest, \Field_Validation $fieldRequest): void
    {
        $cartRequest->setMethod('eid', 'required');
        $cartRequest->setMethod('quantity', 'required');
        $cartRequest->setMethod('quantity', 'regex', '@^\d+$@'); // 整数
        $cartRequest->setMethod('quantity', 'min', 1);

        $cartRequest->validate(new ACMS_Validator());
        $fieldRequest->validate(new ACMS_Validator()); // カートフィールド
    }
}
