<?php

namespace Acms\Plugins\ShoppingCart\POST\ShoppingCart\Update;

use Acms\Plugins\ShoppingCart\POST\ShoppingCart;
use ACMS_Validator;
use ACMS_POST_Cache;
use Field;
use AcmsLogger;

/**
 * カートアイテムの数量を更新する
 */
class Quantity extends ShoppingCart
{
    /**
     * @return Field_Validation|never
     */
    public function post()
    {
        $this->setBlock('update.quantity');

        $cartRequest = $this->extract('cart');

        $this->validateRequest($cartRequest);

        if (!$this->Post->isValidAll()) {
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice('【Shopping Cart plugin】バリデートに失敗したため、カート商品の数量更新に失敗しました', [
                    'cart' => $cartRequest->_aryV,
                ]);
            }
            return $this->Post;
        }

        $cart = $this->loadCart();

        $entryId = idval($cartRequest->get('eid'));
        $quantity = intval($cartRequest->get('quantity'));

        $item = $cart->getItemByEntryId($entryId);

        if (is_null($item)) {
            $cartRequest->setValidator('eid', 'exist', false);
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice(
                    '【Shopping Cart plugin】指定したエントリーIDの商品がカート内に存在しないため、カート商品の数量更新に失敗しました',
                    [
                        'entryId' => $entryId,
                    ]
                );
            }
            return $this->Post;
        }

        $item->setQuantity($quantity);

        // 在庫チェック
        if ($item->isStockManagementEnabled() && $item->isStockShortage()) {
            $cartRequest->setValidator('quantity', 'stock', false);
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice(
                    '【Shopping Cart plugin】数量が在庫数を超過しているため、カート商品（' . $item->getEntry()->getTitle() . '）の数量更新に失敗しました',
                    [
                        'entryId' => $entryId,
                        'quantity' => $quantity,
                        'stock' => intval($item->getEntry()->getField()->get('item_stock')),
                    ]
                );
            }
            return $this->Post;
        }

        $this->saveCart($cart);

        if (class_exists('AcmsLogger')) {
            AcmsLogger::info('【Shopping Cart plugin】カート内商品（' . $item->getEntry()->getTitle() . '）の数量更新に成功しました', [
                'entryId' => $entryId,
                'quantity' => $quantity,
            ]);
        }

        $this->setCartItem($item);

        if (config('cache_clear_when_post') === 'on') {
            ACMS_POST_Cache::clearPageCache(BID);
        }

        $this->redirect(HTTP_REQUEST_URL);
    }

    /**
     * 入力値のバリデート
     *
     * @param \Field_Validation $cartRequest
     * @return void
     */
    protected function validateRequest(\Field_Validation $cartRequest): void
    {
        $cartRequest->setMethod('eid', 'required');
        $cartRequest->setMethod('quantity', 'required');
        $cartRequest->setMethod('quantity', 'regex', '@^\d+$@'); // 整数
        $cartRequest->setMethod('quantity', 'min', 1);

        $cartRequest->validate(new ACMS_Validator());
    }
}
