<?php

namespace Acms\Plugins\ShoppingCart\POST\ShoppingForm;

use Acms\Plugins\ShoppingCart\Entities\Cart;
use Acms\Plugins\ShoppingCart\Entities\Item;
use Acms\Plugins\ShoppingCart\Entities\Order;
use Acms\Plugins\ShoppingCart\Repositories\CartRepository;
use Acms\Plugins\ShoppingCart\Services\ShoppingService;
use App;
use Field;
use ACMS_POST_Form_Confirm;
use AcmsLogger;

/**
 * 決済フォームの確認
 */
class Confirm extends ACMS_POST_Form_Confirm
{
    /**
     * @var CartRepository
     */
    protected $cartRepository;

    /**
     * @var ShoppingService
     */
    protected $ShoppingService;

    /**
     * @var Order|null
     */
    protected $Order;

    /**
     * @var Cart|null
     */
    protected $Cart;

    /**
     * constructor
     */
    public function __construct()
    {
        $this->cartRepository = App::make('shopping-cart.repository.cart');
        $this->ShoppingService = App::make('shopping-cart.service.shopping');
    }

    /**
     * ready
     */
    protected function ready()
    {
        $this->Cart = $this->cartRepository->find(BID);
    }

    /**
     * Run
     *
     * @return Field_Validation
     */
    public function post()
    {
        parent::post();

        /**
         * Ready
         */
        $this->ready();

        /**
         * Set Order
         */
        $this->setOrder();

        if (is_null($this->Order)) {
            // 購入処理中の受注情報がないためエラー
            $this->Post->set('step', 'forbidden');
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice('【Shopping Cart plugin】購入処理中の注文情報が存在しないため、購入処理を中断しました。');
            } else {
                userErrorLog('ACMS Shopping Error: In ShoppingCart extension -> Not Found Order');
            }
            return $this->Post;
        }

        if (!$this->validate()) {
            $this->Post->set('step', 'shoppingError');
            if (class_exists('AcmsLogger')) {
                // ToDo: より詳細に情報を出せるよう検討
                AcmsLogger::notice('【Shopping Cart plugin】注文情報のバリデートに失敗したため、購入処理を中断しました。');
            }
            return $this->Post;
        }

        return $this->Post;
    }

    /**
     * 注文商品の取得
     *
     * @return Item[]
     */
    protected function getItems(): array
    {
        return $this->Cart->getItems();
    }

    /**
     * 受注処理中の注文情報IDの取得
     *
     * @return string|null
     */
    protected function getOrderProcessingId(): ?string
    {
        return $this->Cart->getOrderProcessingId();
    }

    /**
     * 注文情報のセット
     *
     * @return void
     */
    protected function setOrder(): void
    {
        // 購入プロセス中の注文情報を取得
        // 冪等性（https://developer.squareup.com/docs/build-basics/common-api-patterns/idempotency）のため、
        // 注文IDを同一購入プロセスでは使い回す必要がある。
        $Order = $this->loadProcessingOrder();

        // 購入プロセス中の注文情報がnullの場合は不正判定なので、注文情報を更新しない。
        if (!is_null($Order)) {
            // 注文情報を最新の状態に更新
            $Order = $this->ShoppingService->updateOrder($Order, $this->getItems(), $this->Post->getChild('field'));
            $this->saveProcessingOrder($Order);
        }

        $this->Order = $Order;
    }

    /**
     * 購入処理中の注文情報を取得する
     *
     * @return Order|null
     */
    public function loadProcessingOrder(): ?Order
    {
        $orderProcessingId = $this->getOrderProcessingId();
        if (isset($orderProcessingId) && $orderProcessingId !== '') {
            return $this->ShoppingService->loadProcessingOrder($orderProcessingId);
        }

        return null;
    }

    /**
     * 購入処理中の注文情報を保存する
     *
     * @param Order $Order
     * @return void
     */
    protected function saveProcessingOrder(Order $Order): void
    {
        $this->ShoppingService->saveProcessingOrder($Order);
        // GETモジュールで受注処理中の注文情報を取得できるように$Postに保存
        $this->Post->set('processingOrderId', $Order->getProcessingId());
    }

    /**
     * 注文情報のバリデート
     *
     * @return bool
     */
    protected function validate(): bool
    {
        return $this->ShoppingService->validate($this->Order);
    }
}
