<?php

namespace Acms\Plugins\ShoppingCart\POST\ShoppingForm;

use Acms\Plugins\ShoppingCart\Entities\Cart;
use Acms\Plugins\ShoppingCart\Entities\Item;
use Acms\Plugins\ShoppingCart\Entities\Order;
use Acms\Plugins\ShoppingCart\Entities\Addition;
use Acms\Plugins\ShoppingCart\Exceptions\ShoppingException;
use Acms\Plugins\ShoppingCart\Repositories\CartRepository;
use Acms\Plugins\ShoppingCart\Services\ShoppingService;
use Acms\Plugins\ShoppingCart\Services\Payment\PaymentMethodInterface;
use Acms\Plugins\ShoppingCart\Services\Payment\PaymentResult;
use App;
use Field;
use Field_Validation;
use ACMS_POST_Form_Submit;
use ACMS_Validator;
use DB;
use AcmsLogger;
use Common;

/**
 * 決済フォームの送信
 */
class Submit extends ACMS_POST_Form_Submit
{
    var $isCacheDelete = true; // phpcs:ignore

    /**
     * @var CartRepository
     */
    protected $cartRepository;

    /**
     * @var ShoppingService
     */
    protected $ShoppingService;

    /**
     * @var Order|null
     */
    protected $Order;

    /**
     * @var Cart|null
     */
    protected $Cart;

    /**
     * constructor
     */
    public function __construct()
    {
        $this->cartRepository = App::make('shopping-cart.repository.cart');
        $this->ShoppingService = App::make('shopping-cart.service.shopping');
    }

    /**
     * ready
     */
    protected function ready()
    {
        $this->Cart = $this->cartRepository->find(BID);
    }

    /**
     * Run
     *
     * @return Field_Validation
     * @throws Exception
     */
    public function post()
    {
        // フォーム情報のロード
        $id = $this->Post->get('id');
        $info = $this->loadForm($id);
        if (empty($info)) {
            if (class_exists('AcmsLogger')) {
                AcmsLogger::critical('【Shopping Cart plugin】フォームID「' . $id . '」が存在しないため、フォーム送信の処理を中断しました');
            } else {
                userErrorLog('Not Exists Form ID.');
            }
            $this->Post->set('step', 'forbidden');
            return $this->Post;
        }

        if (config('form_csrf_enable', 'on') !== 'off') {
            // CSRF対策
            if (!$this->checkCsrfToken()) {
                if (class_exists('AcmsLogger')) {
                    AcmsLogger::notice('【Shopping Cart plugin】フォームID「' . $id . '」で、CSRFトークンが一致しないため、処理を中断しました');
                }
                $this->Post->set('step', 'forbidden');
                return $this->Post;
            }
            // 2重送信対策
            if (!$this->checkDoubleSubmit()) {
                if (class_exists('AcmsLogger')) {
                    AcmsLogger::notice('【Shopping Cart plugin】フォームID「' . $id . '」で、二重送信を検知したため処理を中断しました');
                }
                $this->Post->set('step', 'repeated');
                return $this->Post;
            }
        }
        if (isCSRF()) {
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice('【Shopping Cart plugin】フォームID「' . $id . '」で、Referrerから外部からのリクエストと判断されたため処理を中断しました');
            }
            $this->Post->set('step', 'forbidden');
            return $this->Post;
        }

        $Form = $info['data'];
        $Mail =& $this->mergeMainConfig($Form->getChild('mail'));

        // サーバサイドのバリデーションを実装
        $Option = new Field();
        $Option->overload($Form->getChild('option'));
        $dup = $this->buildOptions($Option);

        // POSTの整形
        $Field = $this->extract('field');
        if (!empty($dup)) {
            list($fd, $mail) = $dup;
            $Field->setMethod($fd, 'duplication', $this->mailToDouble($info['id'], $mail));
        }

        // バリデート
        $Field->validate(new ACMS_Validator());
        if (!$this->Post->isValidAll()) {
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice('【Shopping Cart plugin】フォームID「' . $id . '」で、バリデートに失敗したため、送信処理を中断しました');
            }
            return $this->Post;
        }

        /**
         * Ready
         */
        $this->ready();

        /**
         * Set Order
         */
        $this->setOrder();

        if (is_null($this->Order)) {
            // 購入処理中の注文情報がないためエラー
            $this->Post->set('step', 'forbidden');
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice('【Shopping Cart plugin】購入処理中の注文情報が存在しないため、購入処理を中断しました。');
            } else {
                userErrorLog('ACMS Shopping Error: In ShoppingCart extension -> Not Found Order');
            }
            return $this->Post;
        }

        if (!$this->validate()) {
            $this->Post->set('step', 'shoppingError');
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice('【Shopping Cart plugin】注文情報のバリデートに失敗したため、購入処理を中断しました。');
            }
            return $this->Post;
        }

        $DB = DB::singleton(dsn());
        $DB->connection()->beginTransaction();

        try {
            // 決済処理
            $PaymentMethod = $this->createPaymentMethod();
            $PaymentResult = $this->execCheckout($PaymentMethod, $DB);

            // 決済が失敗した場合
            if (!$PaymentResult->isSuccess()) {
                $DB->connection()->rollBack();
                if (class_exists('AcmsLogger')) {
                    AcmsLogger::notice('【Shopping Cart plugin】決済処理中にエラーが発生したため、決済処理を中断しました', [
                        'paymentMethod' => $PaymentMethod->getCode(),
                        'errors' => $PaymentResult->getErrors(),
                        'message' => $PaymentResult->getMessage(),
                        'details' => $PaymentResult->getDetails(),
                    ]);
                } else {
                    userErrorLog('ACMS Payment Error: In ShoppingCart extension -> ' . $PaymentResult->getMessage());
                }
                return $this->Post;
            }

            $DB->connection()->commit();
        } catch (ShoppingException $e) {
            $DB->connection()->rollBack();
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice('【Shopping Cart plugin】購入処理中にエラーが発生したため、購入処理を中断しました', Common::exceptionArray($e));
            } else {
                userErrorLog('ACMS Shopping Error: In ShoppingCart extension -> ' . $e->getMessage());
            }
            $this->Post->set('step', 'shoppingError');
            return $this->Post;
        } catch (\Exception $e) {
            $DB->connection()->rollBack();
            if (class_exists('AcmsLogger')) {
                AcmsLogger::notice('【Shopping Cart plugin】購入処理中に不明なエラーが発生したため、購入処理を中断しました', Common::exceptionArray($e));
            } else {
                userErrorLog('ACMS Unknown Error: In ShoppingCart extension -> ' . $e->getMessage());
            }
            $this->Post->set('step', 'shoppingError');
            return $this->Post;
        }

        // メール送信はトランザクションに含めない
        // 決済処理は成功したけど、メール送信が失敗した場合にrollBackしてしまうと
        // 決済サービス側では成功したことになるが、a-blog cms側では失敗したことになってしまうため
        try {
            // 連番の更新
            $this->updateCount($info['id'], $Field);

            // メール用変数の追加
            $this->setMailVars($this->Order, $Field);

            // メール送信
            $admin_log = $this->sendToAdministrator($Mail, $Field);
            $log = $this->sendAutoReply($Mail, $Field);
        } catch (\Exception $e) {
            $this->Post->set('step', 'forbidden');
            if (class_exists('AcmsLogger')) {
                AcmsLogger::warning(
                    '【Shopping Cart plugin】フォームID「' . $id . '」でメール送信に失敗しました',
                    Common::exceptionArray($e)
                );
            } elseif (DEBUG_MODE) {
                throw $e;
            }

            return $this->Post;
        }

        try {
            // ログを記憶
            if (!(isset($info['log']) && $info['log'] === '0')) {
                $this->updateLog($info, $log, $admin_log, $Field);
            }
        } catch (\Exception $e) {
            $this->Post->set('step', 'forbidden');
            if (class_exists('AcmsLogger')) {
                AcmsLogger::warning(
                    '【Shopping Cart plugin】フォームID「' . $id . '」で送信結果のDB保存に失敗しました',
                    Common::exceptionArray($e)
                );
            } elseif (DEBUG_MODE) {
                throw $e;
            }
            return $this->Post;
        }

        // セッション管理のデータを削除を登録
        $this->clearSession();

        return $this->Post;
    }

    /**
     * 注文商品の取得
     *
     * @return Item[]
     */
    protected function getItems(): array
    {
        return $this->Cart->getItems();
    }

    /**
     * 受注処理中の注文情報IDの取得
     *
     * @return string|null
     */
    protected function getOrderProcessingId(): ?string
    {
        return $this->Cart->getOrderProcessingId();
    }

    /**
     * 注文情報のセット
     *
     * @return void
     */
    protected function setOrder(): void
    {
        // 購入プロセス中の注文情報を取得
        // 冪等性（https://developer.squareup.com/docs/build-basics/common-api-patterns/idempotency）のため、
        // 注文IDを同一購入プロセスでは使い回す必要がある。
        $Order = $this->loadProcessingOrder();

        // 購入プロセス中の注文情報がnullの場合は不正判定なので、注文情報を更新しない。
        if (!is_null($Order)) {
            // 注文情報を最新の状態に更新
            $Order = $this->ShoppingService->updateOrder($Order, $this->getItems(), $this->Post->getChild('field'));
            $this->saveProcessingOrder($Order);
        }

        $this->Order = $Order;
    }

    /**
     * 購入処理中の注文情報を取得する
     *
     * @return Order|null
     */
    public function loadProcessingOrder(): ?Order
    {
        $orderProcessingId = $this->getOrderProcessingId();
        if (isset($orderProcessingId) && $orderProcessingId !== '') {
            return $this->ShoppingService->loadProcessingOrder($orderProcessingId);
        }

        return null;
    }

    /**
     * 購入処理中の注文情報を保存する
     *
     * @param Order $Order
     * @return void
     */
    protected function saveProcessingOrder(Order $Order): void
    {
        $this->ShoppingService->saveProcessingOrder($Order);
        $this->saveOrderProcessingId($Order->getProcessingId());
        // GETモジュールで受注処理中の注文情報を取得できるように$Postに保存
        $this->Post->set('processingOrderId', $Order->getProcessingId());
    }

    /**
     * 受注処理中の注文情報IDの保存
     *
     * @param string $orderProcessingId
     * @return void
     */
    protected function saveOrderProcessingId(string $orderProcessingId): void
    {
        $this->Cart->setOrderProcessingId($orderProcessingId);
        $this->cartRepository->save($this->Cart, BID);
    }

    /**
     * 購入処理中の注文情報を削除する
     *
     * @param Order $Order
     * @return void
     */
    protected function deleteProcessingOrder(Order $Order): void
    {
        $this->ShoppingService->deleteProcessingOrder($Order);
    }

    /**
     * 注文情報のバリデート
     *
     *
     * @return bool
     */
    protected function validate(): bool
    {
        return $this->ShoppingService->validate($this->Order);
    }

    /**
     * 決済方法の作成
     *
     * @return PaymentMethodInterface
     */
    protected function createPaymentMethod(): PaymentMethodInterface
    {
        return $this->ShoppingService->createPaymentMethod($this->Order, $this->Post);
    }

    /**
     * 決済処理の実行
     *
     * @param PaymentMethodInterface $PaymentMethod 支払方法
     * @param \Acms\Services\Database\Engine\Base $DB トランザクション制御されているDatabase
     *
     * @return PaymentResult
     */
    protected function execCheckout(
        PaymentMethodInterface $PaymentMethod,
        \Acms\Services\Database\Engine\Base $DB
    ): PaymentResult {
        $PaymentResult = $this->ShoppingService->processCheckout($this->Order, $PaymentMethod, $DB);

        if (!$PaymentResult->isSuccess()) {
            $Field = $this->Post->getChild('field');

            foreach ($PaymentResult->getErrors() as $error) {
                if ($this->Order->getPayment() !== null) {
                    $Field->setValidator($this->Order->getPayment()->getBlock(), $error, false);
                }
            }

            // 決済エラーの場合、idempotency_key（= Order Processing ID） が同一だとエラーになる決済サービス（Square）が存在する。
            // そのため、Order Processing ID の再割当てを行う。
            $this->deleteProcessingOrder($this->Order);
            $this->Order->setProcessingId(uniqid());
            $this->saveProcessingOrder($this->Order);
        }

        return $PaymentResult;
    }


    /**
     * フィールドに追加手数料の変数を追加する
     *
     * @param Order $Order
     * @param Field_Validation $Field
     * @return void
     */
    protected function setMailVars(Order $Order, Field_Validation $Field): void
    {
        // addition
        /** @var Addition[] $Additions */
        $Additions = $Order->getAdditions();
        $data = [
            'id' => [],
            'name' => [],
            'fee' => []
        ];

        foreach ($Additions as $Addition) {
            $data['id'][] = $Addition->getId();
            $data['name'][] = $Addition->getName();
            $data['fee'][] = $Addition->getFee();
        }

        $Field->set('@additionGroup', ['additionId', 'additionName', 'additionFee']);
        foreach ($data as $key => $value) {
            $Field->set('addition' . ucfirst($key), $value);
        }

        // payment
        $Payment = $Order->getPayment();
        if (!is_null($Payment)) {
            $Field->set('paymentMethodBlock', $Payment->getBlock());
            $Field->set('paymentMethodName', $Payment->getName());
            $Field->set('paymentMethodCommissionPaid', $Payment->getCommissionPaid());
            $Field->set('paymentMethodDescription', $Payment->getDescription());
            $Field->set('paymentMethodMailMessage', $Payment->getMailMessage());
        }
    }

    /**
     * セッション管理のデータを削除
     *
     * @return void
     */
    protected function clearSession(): void
    {
        $this->cartRepository->delete(BID);
        register_shutdown_function(function () {
            $this->deleteProcessingOrder($this->Order);
        });
    }
}
