<?php

namespace Acms\Plugins\ShoppingCart\Plugins\Square\Services\OAuth\Contracts;

use DateInterval;
use DateTimeImmutable;
use DateTimeZone;
use Square\SquareClient;
use Square\Apis\OAuthApi;
use Square\Models\ObtainTokenRequest;
use Square\Models\RevokeTokenRequest;
use App;
use DB;
use SQL;
use Cache;
use Common;
use AcmsLogger;
use Acms\Plugins\ShoppingCart\Plugins\Square\Exceptions\SquareOAuthException;
use Acms\Plugins\ShoppingCart\Plugins\Square\Util\SquareUtil;

/**
 * SquareのOAuthServiceの基底クラス
 */
abstract class AbstractSquareOAuthService
{
    /**
     * エンドポイントのベースURL
     *
     * @var string
     */
    protected const BASE_URL = '';

    /**
     * @var \Acms\Plugins\ShoppingCart\Repositories\ConfigRepository
     */
    protected $Config;

    /**
     * アプリケーションで許可するSquare APIの権限
     *
     * @var array
     */
    protected $permissions = [
        'MERCHANT_PROFILE_READ',
        'PAYMENTS_WRITE',
        'ORDERS_WRITE',
        'ORDERS_READ'
    ];

    /**
     * OAuth認証済みか
     *
     * @var bool
     */
    private $isAuthorized = false;

    /**
     * 認証エラー
     *
     * @var bool
     */
    private $isError = false;

    /**
     * アクセストークン
     *
     * @var string|null
     */
    protected $accessToken;

    /**
     * リフレッシュトークン
     *
     * @var string|null
     */
    protected $refreshToken;

    /**
     * アクセストークンの有効期限
     *
     * @var DateTimeImmutable|null
     */
    protected $expiresAt;

    /**
     * アクセストークン取得日
     *
     * @var DateTimeImmutable|null
     */
    protected $obtainedAt;

    /**
     * 販売者ID
     *
     * @var string|null
     */
    protected $merchantId;

    /**
     * Application ID
     *
     * @var string|null
     */
    protected $applicationId;

    /**
     * Application Secret
     *
     * @var string|null
     */
    protected $applicationSecret;

    /**
     * Location ID
     *
     * @var string|null
     */
    protected $locationId;

    /**
     * constractor
     */
    public function __construct(
        ?string $applicationId,
        ?string $applicationSecret
    ) {
        $this->applicationId = $applicationId;
        $this->applicationSecret = $applicationSecret;

        $this->Config = App::make('shopping-cart.repository.config');

        $this->refreshToken = $this->loadRefreshToken();
        $this->expiresAt = $this->loadExpiresAt();
        $this->obtainedAt = $this->loadObtainedAt();
        $this->merchantId = $this->loadMerchantId();

        // データベースからアクセストークンを取得する毎に、アクセストークンの更新が必要かをチェックする
        // see https://developer.squareup.com/docs/oauth-api/refresh-revoke-limit-scope
        if ($this->shouldRefreshAccessToken()) {
            try {
                $this->refreshToken();
            } catch (SquareOAuthException $e) {
                 $this->setIsError(true);
                if (class_exists('AcmsLogger')) {
                    AcmsLogger::error('【Square】 リフレッシュトークンに失敗しました', Common::exceptionArray($e, [
                        'details' => $e->getDetails(),
                    ]));
                } else {
                    userErrorLog('ACMS Error: ' . $e->getMessage());
                }
            }
        }
        $this->accessToken = $this->loadAccessToken();
        $this->locationId = $this->loadLocationId();

        if (!empty($this->accessToken)) {
            $this->isAuthorized = $this->checkAuthorized();

            if (!$this->isAuthorized) {
                $this->removeTokensFromDatabase();
            }
        }
    }

    public function getIsAuthorized(): bool
    {
        return $this->isAuthorized;
    }

    public function setIsAuthorized(bool $isAuthorized): void
    {
        $this->isAuthorized = $isAuthorized;
    }

    public function getIsError(): bool
    {
        return $this->isError;
    }

    public function setIsError(bool $isError): void
    {
        $this->isError = $isError;
    }

    public function getAccessToken(): ?string
    {
        return $this->accessToken;
    }

    public function setAccessToken(?string $accessToken): void
    {
        $this->accessToken = $accessToken;
    }

    public function getRefreshToken(): ?string
    {
        return $this->refreshToken;
    }

    public function setRefreshToken(?string $refreshToken): void
    {
        $this->refreshToken = $refreshToken;
    }

    public function getExpiresAt(): ?DateTimeImmutable
    {
        return $this->expiresAt;
    }

    public function setExpiresAt(?DateTimeImmutable $expiresAt): void
    {
        $this->expiresAt = $expiresAt;
    }

    public function getObtainedAt(): ?DateTimeImmutable
    {
        return $this->obtainedAt;
    }

    public function setObtainedAt(?DateTimeImmutable $obtainedAt): void
    {
        $this->obtainedAt = $obtainedAt;
    }

    public function getMerchantId(): ?string
    {
        return $this->merchantId;
    }

    public function setMerchantId(?string $merchantId): void
    {
        $this->merchantId = $merchantId;
    }

    public function getLocationId(): ?string
    {
        return $this->locationId;
    }

    public function setLocationId(?string $locationId): void
    {
        $this->locationId = $locationId;
    }

    /**
     * Square Location ID を取得する
     *
     * @return string Square Location ID
     */
    abstract protected function loadLocationId(): ?string;

    /**
     * アクセストークンをコンフィグから取得する
     *
     * @return string|null アクセストークン
     */
    abstract protected function loadAccessToken(): ?string;

    /**
     * アクセストークンの取得日時をコンフィグから取得する
     *
     * @return DateTimeImmutable|null アクセストークンの取得日時
     */
    abstract protected function loadObtainedAt(): ?DateTimeImmutable;

    /**
     * アクセストークンの有効期限をコンフィグから取得する
     *
     * @return DateTimeImmutable|null アクセストークンの有効期限
     */
    abstract protected function loadExpiresAt(): ?DateTimeImmutable;

    /**
     * リフレッシュトークンをコンフィグから取得する
     *
     * @return string|null リフレッシュトークン
     */
    abstract protected function loadRefreshToken(): ?string;

    /**
     * Square販売者IDをコンフィグから取得する
     *
     * @return string|null Square販売者ID
     */
    abstract protected function loadMerchantId(): ?string;

    /**
     * アクセストークンを保存する
     *
     * @param string $accessToken
     */
    abstract protected function saveAccessToken(string $accessToken = ''): void;

    /**
     * アクセストークンの有効期限を保存する
     *
     * @param string $expiresAt
     */
    abstract protected function saveExpiresAt(string $expiresAt = ''): void;

    /**
     * アクセストークン取得時間を保存する
     *
     * @param string $obtainedAt
     */
    abstract protected function saveObtainedAt(string $obtainedAt = ''): void;

    /**
     * リフレッシュトークンを保存する
     *
     * @param string $refreshToken
     */
    abstract protected function saveRefreshToken(string $refreshToken = ''): void;

    /**
     * Square販売者IDを保存する
     *
     * @param string $merchantId
     */
    abstract protected function saveMerchantId(string $merchantId = ''): void;

    /**
     * 店舗IDを保存する
     *
     * @param string $locationId
     */
    abstract protected function saveLocationId(string $locationId = ''): void;

    /**
     * APIを使用するモード(サンドボックス or 本番) を取得する
     *
     * @return string APIを使用するモード
     */
    abstract public function getEnvironment(): string;

    /**
     * SquareのOAuthAPIを叩くラッパーを取得する
     *
     * @return OAuthApi
     */
    protected function getOAuthApi(): OAuthApi
    {
        $apiClient = new SquareClient([
            'environment' => $this->getEnvironment()
        ]);
        return $apiClient->getOAuthApi();
    }

    /**
     * SquareのAPIを叩くラッパーを取得する
     *
     * @return SquareClient
     */
    protected function getApiClient(): SquareClient
    {
        return new SquareClient([
            'accessToken' => $this->getAccessToken(),
            'environment' => $this->getEnvironment()
        ]);
    }


    /**
     * アクセストークンの更新が必要かどうかを判定する
     *
     * @return bool
     */
    public function shouldRefreshAccessToken(): bool
    {
        $obtainedAt = $this->getObtainedAt();

        if (is_null($obtainedAt)) {
            return false;
        }

        if ($obtainedAt->add(new DateInterval('P7D')) < new DateTimeImmutable('now')) {
            return true;
        }

        // アクセストークンの取得日から7日以内の場合でも、
        // 念の為トークンの有効期限が切れていないかどうかを確認
        $expiresAt = $this->getExpiresAt();

        if (is_null($expiresAt)) {
            return false;
        }

        // -1日にしているのは、処理の途中で有効期限が切れることがないようにするため
        if ($expiresAt->sub(new DateInterval('P1D')) < new DateTimeImmutable('now', new DateTimeZone('Z'))) {
            return true;
        }

        return false;
    }

    /**
     * データベースに保存されているSquare認証情報を削除する
     */
    protected function removeTokensFromDatabase()
    {
        if (!empty($this->getAccessToken())) {
            $this->saveAccessToken('');
        }
        if (!empty($this->getRefreshToken())) {
            $this->saveRefreshToken('');
        }
        if (!empty($this->getExpiresAt())) {
            $this->saveExpiresAt('');
        }
        if (!empty($this->getMerchantId())) {
            $this->saveMerchantId('');
        }
        if (!empty($this->getLocationId())) {
            $this->saveLocationId('');
        }
        if (!empty($this->getObtainedAt())) {
            $this->saveObtainedAt('');
        }
    }

    /**
     * configテーブルの要素を保存する
     *
     * @param string $key
     * @param string $value
     */
    protected function saveConfig($key, $value)
    {
        Cache::flush('config');

        $DB = DB::singleton(dsn());
        $RemoveSQL = SQL::newDelete('config');
        $RemoveSQL->addWhereOpr('config_blog_id', BID);
        $RemoveSQL->addWhereOpr('config_key', $key);
        $DB->query($RemoveSQL->get(dsn()), 'exec');

        $InsertSQL = SQL::newInsert('config');
        $InsertSQL->addInsert('config_key', $key);
        $InsertSQL->addInsert('config_value', $value);
        $InsertSQL->addInsert('config_blog_id', BID);
        $DB->query($InsertSQL->get(dsn()), 'exec');
    }

    /**
     * アプリケーションが認証済みかどうかを調査する
     *
     * @return bool
     */
    public function checkAuthorized(): bool
    {
        // 店舗情報を取得できる場合、認証済み
        try {
            $this->getLocations();
            return true;
        } catch (\Exception $e) {
            return false;
        }
    }

    /**
     * Squareから認証コードを取得するためのURLを作成する
     *
     * @param string $state コールバックのgetパラメタとして渡すstate(csrf対策として、ブラウザを一意に対応付けられる値)
     * @return string URL
     */
    public function buildUrlGettingAuthorizationCode($state)
    {
        $permissions = urlencode(
            implode(' ', $this->permissions)
        );

        return static::BASE_URL . '/oauth2/authorize' .
            '?client_id=' . $this->applicationId .
            '&scope=' . $permissions .
            '&state=' . $state .
            '&session=false';
    }

    /**
     * Squareからトークンを取得し、更新する
     *
     * @param string $authorizationCode コールバックのgetパラメタとして渡されるcode
     * @param string $parameterState コールバックのgetパラメタとして渡されるstate
     * @param string $comparedState csrf対策のstate
     *
     * @throws SquareOAuthException アクセストークンの取得に失敗した場合に投げられる
     */
    public function obtainToken($authorizationCode, $parameterState, $comparedState)
    {
        // cross-site request forgery対策
        if ($parameterState !== $comparedState) {
            throw new SquareOAuthException('Error: parameter state is different.');
        }

        $body = new ObtainTokenRequest(
            $this->applicationId,
            'authorization_code'
        );

        $body->setClientSecret($this->applicationSecret);
        $body->setCode($authorizationCode);

        // OAuth APIを叩く
        $response = $this->getOAuthApi()->obtainToken($body);

        if ($response->isError()) {
            throw new SquareOAuthException(
                SquareUtil::parseResponseError(__FUNCTION__, $response->getErrors()),
                json_decode(json_encode($response->getErrors()), true)
            );
        }

        $accessToken = $response->getResult()->getAccessToken();
        $expiresAt = $response->getResult()->getExpiresAt();
        $refreshToken = $response->getResult()->getRefreshToken();
        $merchantId = $response->getResult()->getMerchantId();
        $obtainedAt = new DateTimeImmutable('now');

        $this->saveAccessToken($accessToken);
        $this->saveExpiresAt($expiresAt);
        $this->saveRefreshToken($refreshToken);
        $this->saveMerchantId($merchantId);
        $this->saveObtainedAt($obtainedAt->format(DateTimeImmutable::ISO8601));

        $this->setAccessToken($accessToken);
        $this->setExpiresAt(new DateTimeImmutable($expiresAt));
        $this->setRefreshToken($refreshToken);
        $this->setMerchantId($merchantId);
        $this->setObtainedAt($obtainedAt);

        $this->setIsAuthorized(true);
    }

    /**
     * Squareから取得したトークンを破棄する
     *
     * @throws SquareOAuthException アクセストークンの削除に失敗した場合に投げられる
     */
    public function revokeToken()
    {
        $body = new RevokeTokenRequest();
        $body->setClientId($this->applicationId);
        $body->setAccessToken($this->getAccessToken());
        $body->setRevokeOnlyAccessToken(false);

        // OAuth APIを叩く
        $response = $this->getOAuthApi()->revokeToken($body, 'Client ' . $this->applicationSecret);

        if ($response->isError()) {
            throw new SquareOAuthException(
                SquareUtil::parseResponseError(__FUNCTION__, $response->getErrors()),
                json_decode(json_encode($response->getErrors()), true)
            );
        }

        $this->removeTokensFromDatabase();

        $this->setAccessToken(null);
        $this->setExpiresAt(null);
        $this->setRefreshToken(null);
        $this->setMerchantId(null);
        $this->setObtainedAt(null);
        $this->setLocationId(null);

        $this->setIsAuthorized(false);
    }

    /**
     * Squareからトークンを再取得する
     *
     * @throws SquareOAuthException アクセストークンの取得に失敗した場合に投げられる
     */
    public function refreshToken()
    {
        $body = new ObtainTokenRequest(
            $this->applicationId,
            'refresh_token'
        );

        $body->setClientSecret($this->applicationSecret);
        $body->setRefreshToken($this->getRefreshToken());

        // OAuth APIを叩く
        $response = $this->getOAuthApi()->obtainToken($body);

        if ($response->isError()) {
            throw new SquareOAuthException(
                SquareUtil::parseResponseError(__FUNCTION__, $response->getErrors()),
                json_decode(json_encode($response->getErrors()), true)
            );
        }

        $accessToken = $response->getResult()->getAccessToken();
        $expiresAt = $response->getResult()->getExpiresAt();
        $refreshToken = $response->getResult()->getRefreshToken();
        $merchantId = $response->getResult()->getMerchantId();
        $obtainedAt = new DateTimeImmutable('now');

        $this->saveAccessToken($accessToken);
        $this->saveExpiresAt($expiresAt);
        $this->saveRefreshToken($refreshToken);
        $this->saveMerchantId($merchantId);
        $this->saveObtainedAt($obtainedAt->format(DateTimeImmutable::ISO8601));

        $this->setAccessToken($accessToken);
        $this->setExpiresAt(new DateTimeImmutable($expiresAt));
        $this->setRefreshToken($refreshToken);
        $this->setMerchantId($merchantId);
        $this->setObtainedAt($obtainedAt);
    }

    /**
     * Location 情報を取得する
     *
     * @return \Square\Models\Location[] Location情報のオブジェクトのリスト
     * @throws SquareOAuthException Location IDの取得に失敗した場合に投げられる
     */
    public function getLocations()
    {
        $locationApi = $this->getApiClient()->getLocationsApi();

        $apiResponse = $locationApi->listLocations();

        if ($apiResponse->isError()) {
            throw new SquareOAuthException(
                SquareUtil::parseResponseError(__FUNCTION__, $apiResponse->getErrors()),
                json_decode(json_encode($apiResponse->getErrors()), true)
            );
        }

        return $apiResponse->getResult()->getLocations();
    }
}
