<?php

namespace Acms\Plugins\ShoppingCart\Plugins\Square\Services\OAuth;

use Square\Environment;
use DateTimeImmutable;
use App;
use Common;
use Storage;
use Acms\Plugins\ShoppingCart\Plugins\Square\Services\OAuth\Contracts\AbstractSquareOAuthService;

/**
 * SquareのSandbox OAuth認証用サービス
 */
class SquareSandboxOAuthService extends AbstractSquareOAuthService
{
    /**
     * {@inheritdoc}
     */
    protected const BASE_URL = 'https://connect.squareupsandbox.com';

    /**
     * @var \Acms\Plugins\ShoppingCart\Repositories\ConfigRepository
     */
    protected $Config;

    /**
     * Square Location ID を取得する
     *
     * @return string|null Square Location ID
     */
    protected function loadLocationId(): ?string
    {
        return $this->Config->get('SandboxSquareLocationID');
    }

    /**
     * アクセストークンをコンフィグから取得する
     *
     * @return string|null アクセストークン
     */
    protected function loadAccessToken(): ?string
    {
        $encrypted = $this->Config->get('sandbox_square_oauth_accesstoken', '');

        if ($encrypted === '') {
            return null;
        }

        $encripter = App::make('shopping-cart.service.encrypter');
        return $encripter->decrypt($encrypted, env('SANDBOX_SQUARE_OAUTH_ACCESSTOKEN_PASSPHRASE_BID_' . BID)) ?: null;
    }

    /**
     * アクセストークンの取得日時をコンフィグから取得する
     *
     * @return DateTimeImmutable|null アクセストークンの取得日時
     */
    protected function loadObtainedAt(): ?DateTimeImmutable
    {
        $obtainedAt = $this->Config->get('sandbox_square_oauth_accesstoken_obtainedat');
        return !empty($obtainedAt) ? new DateTimeImmutable($obtainedAt) : null;
    }

    /**
     * アクセストークンの有効期限をコンフィグから取得する
     *
     * @return DateTimeImmutable|null アクセストークンの有効期限
     */
    protected function loadExpiresAt(): ?DateTimeImmutable
    {
        $expiresAt = $this->Config->get('sandbox_square_oauth_accesstoken_expiresat');
        return !empty($expiresAt) ? new DateTimeImmutable($expiresAt) : null;
    }

    /**
     * リフレッシュトークンをコンフィグから取得する
     *
     * @return string|null リフレッシュトークン
     */
    protected function loadRefreshToken(): ?string
    {
        $encrypted = $this->Config->get('sandbox_square_oauth_refreshtoken', '');

        if ($encrypted === '') {
            return null;
        }

        $encripter = App::make('shopping-cart.service.encrypter');
        return $encripter->decrypt($encrypted, env('SANDBOX_SQUARE_OAUTH_REFRESHTOKEN_PASSPHRASE_BID_' . BID)) ?: null;
    }

    /**
     * Square販売者ID をコンフィグから取得する
     *
     * @return string|null Square販売者ID
     */
    protected function loadMerchantId(): ?string
    {
        return $this->Config->get('sandbox_square_oauth_merchantid');
    }

    /**
     * APIを使用するモード(サンドボックス or 本番) を取得する
     *
     * @return string APIを使用するモード
     */
    public function getEnvironment(): string
    {
        return Environment::SANDBOX;
    }

    /**
     * アクセストークンを保存する
     *
     * @param string $accessToken
     */
    protected function saveAccessToken(string $accessToken = ''): void
    {
        if ($accessToken === '') {
            $this->saveConfig('sandbox_square_oauth_accesstoken', '');
            return;
        }

        $encripter = App::make('shopping-cart.service.encrypter');

        $key = 'SANDBOX_SQUARE_OAUTH_ACCESSTOKEN_PASSPHRASE_BID_' . BID;

        $passphrase = env($key);

        // パスフレーズは環境変数として管理するが、
        // サイト制作者は意識しなくても良いように自動生成を行う。
        if (empty($passphrase)) {
            $env = SCRIPT_DIR . '.env';
            if (!Storage::exists($env)) {
                throw new \RuntimeException('Not Found .env');
            }
            $passphrase = Common::genPass(8);
            Storage::put($env, Storage::get($env) . $key . '=' . $passphrase . PHP_EOL);
        }

        $encrypted = $encripter->encrypt($accessToken, $passphrase);
        $this->saveConfig('sandbox_square_oauth_accesstoken', $encrypted);
    }

    /**
     * アクセストークンの有効期限を保存する
     *
     * @param string $expiresAt
     */
    protected function saveExpiresAt(string $expiresAt = ''): void
    {
        $this->saveConfig('sandbox_square_oauth_accesstoken_expiresat', $expiresAt);
    }

    /**
     * アクセストークン取得時間を更新する
     */
    protected function saveObtainedAt(string $obtainedAt = ''): void
    {
        $this->saveConfig('sandbox_square_oauth_accesstoken_obtainedat', $obtainedAt);
    }

    /**
     * リフレッシュトークンを保存する
     *
     * @param string $refreshToken
     */
    protected function saveRefreshToken(string $refreshToken = ''): void
    {
        if ($refreshToken === '') {
            $this->saveConfig('sandbox_square_oauth_refreshtoken', '');
            return;
        }

        $encripter = App::make('shopping-cart.service.encrypter');

        $key = 'SANDBOX_SQUARE_OAUTH_REFRESHTOKEN_PASSPHRASE_BID_' . BID;

        $passphrase = env($key);

        // パスフレーズは環境変数として管理するが、
        // サイト制作者は意識しなくても良いように自動生成を行う。
        if (empty($passphrase)) {
            $env = SCRIPT_DIR . '.env';
            if (!Storage::exists($env)) {
                throw new \RuntimeException('Not Found .env');
            }
            $passphrase = Common::genPass(8);
            Storage::put($env, Storage::get($env) . $key . '=' . $passphrase . PHP_EOL);
        }

        $encrypted = $encripter->encrypt($refreshToken, $passphrase);
        $this->saveConfig('sandbox_square_oauth_refreshtoken', $encrypted);
    }

    /**
     * Square販売者IDを保存する
     *
     * @param string $merchantId
     */
    protected function saveMerchantId(string $merchantId = ''): void
    {
        $this->saveConfig('sandbox_square_oauth_merchantid', $merchantId);
    }

    /**
     * 店舗IDを保存する
     */
    protected function saveLocationId(string $locationId = ''): void
    {
        $this->saveConfig('SandboxSquareLocationID', $locationId);
    }
}
