<?php

namespace Acms\Plugins\ShoppingCart\Plugins\Square\Services;

use Square\SquareClient;
use Square\Models\CreatePaymentRequest;
use Square\Models\CreateOrderRequest;
use Square\Models\OrderLineItem;
use Square\Models\OrderLineItemTax;
use Square\Models\OrderLineItemTaxType;
use Square\Models\OrderLineItemTaxScope;
use Square\Models\OrderServiceChargeCalculationPhase;
use Square\Models\OrderServiceCharge;
use Square\Models\OrderLineItemPricingBlocklists;
use Square\Models\OrderLineItemPricingBlocklistsBlockedTax;
use Square\Models\Money;
use Square\Models\Currency;
use Square\Exceptions\ApiException;
use App;
use Acms\Plugins\ShoppingCart\Plugins\Square\Exceptions\SquarePaymentsException;
use Acms\Plugins\ShoppingCart\Plugins\Square\Services\OAuth\Contracts\AbstractSquareOAuthService;
use Acms\Plugins\ShoppingCart\Plugins\Square\Util\SquareUtil;
use Acms\Plugins\ShoppingCart\Entities\Order;
use Acms\Plugins\ShoppingCart\Entities\TaxAmount;
use stdClass;

/**
 * Square側から提供されたintegration IDをCreate paymentのリクエストに含めるためにカスタマイズしたCreatePaymentRequest
 * appleple経由での決済であることの証明であり、integration IDを含んだリクエストで決済することでapplepleの利益になる
 */
class CustomCreatePaymentRequest extends CreatePaymentRequest
{
    /**
     * @var array
    */
    private $auxiliary_info;

    public function __construct(string $sourceId, string $idempotencyKey)
    {
        parent::__construct($sourceId, $idempotencyKey);

        $this->auxiliary_info = [[
            'key' => 'integration_id',
            'value' => 'sqi_19a69bcdb2e04385aabb3fe7f82e238e'
        ]
        ];
    }

    /**
     * Encode this object to JSON
     *
     * @param bool $asArrayWhenEmpty Whether to serialize this model as an array whenever no fields
     *        are set. (default: false)
     *
     * @return array|stdClass
     */
    public function jsonSerialize(bool $asArrayWhenEmpty = false)
    {
        $json = parent::jsonSerialize();
        $json['auxiliary_info'] = $this->auxiliary_info;

        return (!$asArrayWhenEmpty && empty($json)) ? new stdClass() : $json;
    }
}

/**
 * Squareの決済処理用サービス
 */
class SquarePaymentService // phpcs:ignore
{
    /**
     * lineItemのuidのサフィックス
     */
    private const ORDER_LINE_ITEM_TAX_UID_SUFFIX = 'PCTax';

    /**
     * @var AbstractSquareOAuthService
     */
    protected $SquareOAuthService;

    /**
     * SquareのロケーションID
     *
     * @var string|null
     */
    private $locationId;

    /**
     * Square Client
     *
     * @var SquareClient
     */
    private $Client;

    /**
     * @var \Acms\Plugins\ShoppingCart\Repositories\ConfigRepository
     */
    protected $Config;

    public function __construct()
    {
        $this->Config = App::make('shopping-cart.repository.config');

        $this->SquareOAuthService = App::make('square.service.oauth');
        $this->locationId = $this->SquareOAuthService->getLocationId();

        $this->Client = new SquareClient([
            'accessToken' => $this->SquareOAuthService->getAccessToken(),
            'environment' => $this->SquareOAuthService->getEnvironment()
        ]);
    }

    /**
     * ロケーションIDを取得する
     *
     * @return string|null
     */
    public function getLocationId(): ?string
    {
        return $this->locationId;
    }

    /**
     * 注文アイテムリストを作成する
     *
     * @param Order $Order
     *
     * @return \Square\Models\Order 生成された注文アイテムリスト
     * @throws SquarePaymentsException 注文アイテムリストの作成に失敗した場合に投げられる
     */
    public function createOrder(Order $Order): \Square\Models\Order
    {
        // 設定の取得
        $taxFormat = $this->Config->get('shopping-cart_tax-format', '');
        $taxCalcMethod = $this->Config->get('shopping-cart_tax-calc-method', '');
        $taxRoundingMehtod = $this->Config->get('shopping-cart_tax-rounding', '');
        $locationId = $this->getLocationId();

        if (!in_array($taxFormat, ['intax', 'extax'], true)) {
            throw new SquarePaymentsException('Undefined tax expression format.', ['taxFormat' => $taxFormat]);
        }

        if ($taxCalcMethod === 'pileup') {
            throw new SquarePaymentsException('Pileup tax calc method is not available in Square.');
        }

        if (empty($taxCalcMethod)) {
            throw new SquarePaymentsException('Undefined tax calc method.');
        }

        if ($taxRoundingMehtod === 'round') {
            throw new SquarePaymentsException('Rounding process is not available in Square.');
        }

        if ($taxRoundingMehtod === 'ceil') {
            throw new SquarePaymentsException('Rounding up process is not available in Square.');
        }

        if (empty($taxRoundingMehtod)) {
            throw new SquarePaymentsException('Undefined tax rounding method.');
        }

        if (empty($locationId)) {
            throw new SquarePaymentsException('Not Found Location ID.');
        }

        $body = new CreateOrderRequest();

        $body->setOrder(new \Square\Models\Order($locationId));

        // 税金
        $OrderLineItemTaxes = [];

        $taxRates = array_map(
            function (TaxAmount $TaxAmount) {
                return $TaxAmount->getTaxRate();
            },
            $Order->getTaxAmounts()
        );

        foreach ($taxRates as $taxRate) {
            $taxPercentage = $taxRate * 100;
            $OrderLineItemTax = new OrderLineItemTax();
            // Warn: ここを変えるときは商品の税金部分も変えること。
            $OrderLineItemTax->setUid(strval($taxPercentage) . self::ORDER_LINE_ITEM_TAX_UID_SUFFIX);
            $OrderLineItemTax->setName("{$taxPercentage}%消費税");
            $OrderLineItemTax->setPercentage(strval($taxPercentage));

            // 税種別を設定
            $OrderLineItemTax->setType(
                $taxFormat === 'intax' ? OrderLineItemTaxType::INCLUSIVE : OrderLineItemTaxType::ADDITIVE
            );

            // 税計算方法を設定
            // 税計算方法の違いは、商品側でも考慮する必要がある
            $OrderLineItemTax->setScope(OrderLineItemTaxScope::ORDER);

            array_push($OrderLineItemTaxes, $OrderLineItemTax);
        }

        $body->getOrder()->setTaxes($OrderLineItemTaxes);

        // 商品
        $LineItems = [];

        foreach ($Order->getOrderItems() as $OrderItem) {
            $LineItem = new OrderLineItem($OrderItem->getQuantity());
            $LineItem->setUid($OrderItem->getEntry()->getId());
            $LineItem->setName($OrderItem->getEntry()->getTitle());
            $LineItem->setBasePriceMoney(new Money());
            $LineItem->getBasePriceMoney()->setAmount($OrderItem->getPrice());
            $LineItem->getBasePriceMoney()->setCurrency(Currency::JPY);

            // 通常ORDERスコープの税率は全ての商品に適用されてしまうため、割り戻し方式の場合は、適用しない税率を設定する
            $OrderLineItemPricingBlocklistsBlockedTaxes = [];
            $OrderLineItemTaxes = array_filter(
                $body->getOrder()->getTaxes(),
                function (OrderLineItemTax $OrderLineItemTax) use ($OrderItem) {
                    // 税金UIDに税率の情報を設定しておくことで、税金インスタンスを照らし合わせられるようにしている。
                    return (string)intval(
                        $OrderItem->getTaxRate() * 100
                    ) . self::ORDER_LINE_ITEM_TAX_UID_SUFFIX !== $OrderLineItemTax->getUid(); // 適用しない税率情報を取得
                }
            );

            foreach ($OrderLineItemTaxes as $OrderLineItemTax) {
                $OrderLineItemPricingBlocklistsBlockedTax = new OrderLineItemPricingBlocklistsBlockedTax();
                $OrderLineItemPricingBlocklistsBlockedTax->setTaxUid($OrderLineItemTax->getUid());
                array_push($OrderLineItemPricingBlocklistsBlockedTaxes, $OrderLineItemPricingBlocklistsBlockedTax);
            }

            $OrderLineItemPricingBlocklists = new OrderLineItemPricingBlocklists();
            $OrderLineItemPricingBlocklists->setBlockedTaxes($OrderLineItemPricingBlocklistsBlockedTaxes);
            $LineItem->setPricingBlocklists($OrderLineItemPricingBlocklists);

            array_push($LineItems, $LineItem);
        }
        $body->getOrder()->setLineItems($LineItems);

        // 手数料の追加
        $OrderServiceCharges = [];

        // 決済手数料
        if ($Order->getCommissionPaid() > 0) {
            $CommissionPaidOrderServiceCharge = $this->getOrderServiceCharge(
                'commissionPaid',
                '決済手数料',
                $Order->getCommissionPaid()
            );

            array_push($OrderServiceCharges, $CommissionPaidOrderServiceCharge);
        }

        // 追加手数料
        if ($Order->getAdditionalFee() > 0) {
            $AdditionalFeeOrderServiceCharge = $this->getOrderServiceCharge(
                'additionalFee',
                '追加手数料',
                $Order->getAdditionalFee()
            );

            array_push($OrderServiceCharges, $AdditionalFeeOrderServiceCharge);
        }

        // 送料
        if ($Order->getShippingFee() > 0) {
            $ShippingFeeOrderServiceCharge = $this->getOrderServiceCharge(
                'shippingFee',
                '送料',
                $Order->getShippingFee()
            );

            array_push($OrderServiceCharges, $ShippingFeeOrderServiceCharge);
        }

        $body->getOrder()->setServiceCharges($OrderServiceCharges);

        // Idempotency（冪等性）を担保するためのID
        // see https://developer.squareup.com/docs/build-basics/common-api-patterns/idempotency
        $body->setIdempotencyKey($Order->getProcessingId());

        $apiResponse = $this->Client->getOrdersApi()->createOrder($body);
        if ($apiResponse->isError()) {
            throw new SquarePaymentsException(
                SquareUtil::parseResponseError(__FUNCTION__, $apiResponse->getErrors()),
                json_decode(json_encode($apiResponse->getErrors()), true)
            );
        }

        return $apiResponse->getResult()->getOrder();
    }

    /**
     * 手数料の作成・取得
     *
     * @param string $uid
     * @param string $name
     * @param int $amount
     *
     * @return OrderServiceCharge
     */
    private function getOrderServiceCharge(string $uid, string $name, int $amount): OrderServiceCharge
    {
        $OrderServiceCharge = new OrderServiceCharge();
        $OrderServiceCharge->setUid($uid);
        $OrderServiceCharge->setName($name);

        $Money = new Money();
        $Money->setAmount($amount);
        $Money->setCurrency(Currency::JPY);
        $OrderServiceCharge->setAmountMoney($Money);
        $OrderServiceCharge->setCalculationPhase(OrderServiceChargeCalculationPhase::TOTAL_PHASE);

        return $OrderServiceCharge;
    }

    /**
     * 決済を行う
     *
     * @param int $total 支払い合計金額
     * @param string $orderId 注文アイテムリストのID
     * @param string $nonce トークン化されたクレジット情報
     * @param string $idempotencyKey 冪等性を担保するためのID
     * @param string $note 支払いに関するメモ
     *
     * @throws SquarePaymentsException 支払いに失敗した場合に投げられる
     * @return \Square\Models\Payment
     */
    public function payment(
        int $total,
        string $orderId,
        string $nonce,
        string $idempotencyKey,
        string $note = ''
    ): \Square\Models\Payment {
        $Money = new Money();
        $Money->setAmount($total);
        $Money->setCurrency(Currency::JPY);

        $body = new CustomCreatePaymentRequest($nonce, $idempotencyKey);
        $body->setOrderId($orderId);
        $body->setLocationId($this->getLocationId());
        $body->setAmountMoney($Money);
        $body->setNote($note);

        try {
            $response = $this->Client->getPaymentsApi()->createPayment($body);
        } catch (ApiException $error) {
            throw new SquarePaymentsException($error);
        }

        if ($response->isError()) {
            throw new SquarePaymentsException(
                SquareUtil::parseResponseError(__FUNCTION__, $response->getErrors()),
                json_decode(json_encode($response->getErrors()), true)
            );
        }

        return $response->getResult()->getPayment();
    }
}
