import { appendField, triggerEvent } from '../../util';

import type {
  Card,
  Payments,
  CardInputEventTypes,
  CardInputEvent,
  SqEvent,
  TokenResult,
} from '@square/web-payments-sdk-types';
import type { CardOptions } from './type';

const tokenize = async (paymentMethod: Card): Promise<TokenResult['token']> => {
  const tokenResult = await paymentMethod.tokenize();
  if (tokenResult.status === 'OK') {
    triggerEvent(window, 'acmsSquareTokenizeSuccessed', { detail: { tokenDetails: tokenResult.details } });
    return tokenResult.token;
  }

  triggerEvent(window, 'acmsSquareTokenizeFailed', { detail: { errors: tokenResult.errors } });

  let errorMessage = `Tokenization failed with status: ${tokenResult.status}`;
  if (tokenResult.errors) {
    errorMessage += ` and errors: ${JSON.stringify(tokenResult.errors)}`;
  }

  throw new Error(errorMessage);
};

const processTokenize = async (paymentMethod: Card, submitter: HTMLInputElement | HTMLButtonElement): Promise<void> => {
  const { form } = submitter;

  if (!form) {
    return;
  }

  try {
    // disable the submit button as we await tokenization
    submitter.disabled = true;
    const token = await tokenize(paymentMethod);

    if (token) {
      appendField('square-nonce', token, form);
    }

    submitter.disabled = false;
    if (typeof form.requestSubmit === 'function') {
      return form.requestSubmit(submitter);
    }
    if (form.reportValidity()) {
      const input = document.createElement('input');
      input.type = 'hidden';
      input.name = submitter.name;
      form.appendChild(input);
      form.dispatchEvent(new Event('submit'));
      return form.submit();
    }
  } catch (e) {
    submitter.disabled = false;
    if (e instanceof Error) {
      console.error(e.message); // eslint-disable-line no-console
    }
  }
};

const initCard = async (
  payments: Payments,
  element: HTMLElement,
  options: CardOptions['squareCardOptions'] = {}
): Promise<Card> => {
  const card = await payments.card(options);
  await card.attach(element);

  return card;
};

const defaultOption: CardOptions = {
  submitterElementOrSelector: '[name*="ACMS_POST_ShoppingForm"]',
  condition: (event) => {
    if (!(event.currentTarget instanceof HTMLInputElement || event.currentTarget instanceof HTMLButtonElement)) {
      return false;
    }

    const { form } = event.currentTarget;

    if (!form?.elements.namedItem('payment-method')) {
      return true;
    }

    // 決済方法のフォームと同一ページの場合は、決済フォームがcreditcardのときのみtokenを取得する。
    return (form.elements.namedItem('payment-method') as HTMLInputElement).value === 'creditcard';
  },
  focus: undefined,
  recalculateSize: () => {}, // eslint-disable-line @typescript-eslint/no-empty-function
  squareCardOptions: {
    postalCode: '',
  },
};

export default (payments: Payments) =>
  async (selector: string, options: CardOptions): Promise<void> => {
    const target = document.querySelector<HTMLElement>(selector);

    if (!target) {
      throw new Error('Not found target element');
    }

    const config = { ...defaultOption, ...options };

    const { submitterElementOrSelector, condition, callbacks, focus, recalculateSize, squareCardOptions } = config;

    let card: Card;
    try {
      card = await initCard(payments, target, squareCardOptions);
    } catch (e) {
      console.error('Initializing Card failed', e); // eslint-disable-line no-console
      return;
    }

    if (callbacks) {
      Object.keys(callbacks).forEach((callback) => {
        card?.addEventListener(
          callback as CardInputEventTypes,
          (callbacks as Record<string, (event: SqEvent<CardInputEvent>) => void>)[callback]
        );
      });
    }

    if (focus) {
      if (typeof focus === 'function') {
        focus(async (field) => await card.focus(field)); // eslint-disable-line no-return-await
      } else {
        await card.focus(focus);
      }
    }

    if (recalculateSize) {
      recalculateSize(() => {
        card.recalculateSize();
      });
    }

    const submitter =
      typeof submitterElementOrSelector === 'string'
        ? document.querySelector<HTMLInputElement | HTMLButtonElement>(submitterElementOrSelector)
        : submitterElementOrSelector;

    if (!submitter) {
      throw new Error('Not found submit element');
    }

    const { form } = submitter;

    if (!form) {
      throw new Error('Not found form element');
    }

    const handleClick = async (event: Event) => {
      if (condition(event)) {
        event.preventDefault();
        await processTokenize(card, submitter);
      }
    };

    submitter.addEventListener('click', handleClick);
  };
