import type * as Square from '@square/web-payments-sdk-types';

export type CardOptions = {
  submitterElementOrSelector: HTMLButtonElement | HTMLInputElement | string;
  condition: (event: Event) => boolean;
  callbacks?: {
    /**
     * Callback function that is called when the payment form detected a new
     * likely credit card brand based on the card number.
     */
    cardBrandChanged?(event: Square.SqEvent<Square.CardInputEvent>): void;
    /**
     * Callback function that is called when a form field has an invalid value,
     * and the corresponding error CSS class was added to the element.
     */
    errorClassAdded?(event: Square.SqEvent<Square.CardInputEvent>): void;
    /**
     * Callback function that is called when an invalid value on a form field
     * was corrected, and the corresponding error CSS class was removed from the element
     */
    errorClassRemoved?(event: Square.SqEvent<Square.CardInputEvent>): void;
    /**
     * Callback function that is called when the user pressed the "Escape" key
     * while editing a field.
     */
    escape?(event: Square.SqEvent<Square.CardInputEvent>): void;
    /**
     * Callback function that is called when a form field gained focus, and the
     * corresponding focus CSS class was added to the element.
     */
    focusClassAdded?(event: Square.SqEvent<Square.CardInputEvent>): void;
    /**
     * Callback function that is called when a form field lost focus, and the
     * corresponding focus CSS class was removed from the element.
     */
    focusClassRemoved?(event: Square.SqEvent<Square.CardInputEvent>): void;
    /**
     * Callback function that is called when the current value of the postal
     * code form field changed.
     */
    postalCodeChanged?(event: Square.SqEvent<Square.CardInputEvent>): void;
    /**
     * Callback function that is called when the user has triggered submission
     * of the form by pressing "Enter" while editing a field.
     */
    submit?(event: Square.SqEvent<Square.CardInputEvent>): void;
  };
  /**
   * Sets the DOM focus of one of the input fields within the credit card form.
   *
   * For more details about the available options, see
   * [CardFieldNames](https://developer.squareup.com/reference/sdks/web/payments/enums/CardFieldNames).
   *
   * @throws {InvalidFieldNameError} The specified field name is invalid
   */
  focus?: Square.CardFieldNamesValues | ((callback: Square.Card['focus'] | undefined) => void);
  /**
   * Recalculates the size of the card form.
   *
   * The Card component normally automatically resizes based on the size of the
   * buyer's browser, however if the Card component is contained with an element
   * that has a computed style property of "display: none", then the Card
   * component will no longer have a defined width and therefore will not
   * properly resize between mobile and desktop configurations. Upon being
   * displayed again, the Card component will not automatically update its size
   * to match the browser window.
   *
   * This method recalculateSize() can be used to handle this edge case by
   * forcing the Card component to recalculate its size and display
   * appropriately for mobile or desktop.
   *
   * @example Card.recalculateSize()
   *
   * @throws {PaymentMethodNotAttachedError} `Card` is has not been attached to
   *   a DOM element
   */
  recalculateSize?(callback: Square.Card['recalculateSize'] | undefined): void;
  /**
   * An interface that declares the properties used to customize the appearance
   * and behavior of the card entry form.
   * For more information about applying custom styles to the card form, see the available [CardClassSelectors](https://developer.squareup.com/reference/sdks/web/payments/objects/CardClassSelectors) for styling.
   * @example
   * const cardOptions = getCardOptions();
   * const card = await payments.card(cardOptions);
   *
   * function getCardOptions() {
   *    return {
   *      "postalCode" : "12345",
   *      "style": {
   *        "input": {
   *          "color": "red",
   *        },
   *        // You can use media queries with valid selectors:
   *        "@media screen and (max-width: 600px)": {
   *          "input": {
   *            "fontSize": "12px",
   *          }
   *        }
   *      },
   *    };
   * };
   */
  squareCardOptions?: Square.CardOptions;
};
