<?php

namespace Acms\Plugins\ShoppingCart\Repositories;

use Acms\Services\Cache\Contracts\AdapterInterface;
use App;
use Session;
use Acms\Services\Facades\Cache;
use Acms\Plugins\ShoppingCart\Entities\Cart;
use Acms\Plugins\ShoppingCart\Entities\Entry;
use Acms\Plugins\ShoppingCart\Entities\Item;
use Acms\Plugins\ShoppingCart\Repositories\EntryRepository;

class CartRepository
{
    /**
     * カート内容を保存するために使用するセッションキーのプレフィックス
     *
     * @var string
     */
    private const SESSION_PREFIX = 'acms_cart_';

    /**
     * カート内容を保存するため使用するキャッシュキーのプレフィックス
     *
     * @var string
     */
    private const CACHE_PREFIX = 'acms-cart-';

    protected EntryRepository $EntryRepository;

    protected AdapterInterface $cache;

    /**
     * constractor
     */
    public function __construct()
    {
        $this->EntryRepository = App::make('shopping-cart.repository.entry');
        $this->cache = Cache::temp();
    }

    /**
     * カートの保存する
     *
     * @param Cart $cart
     * @param int $bid
     */
    public function save(Cart $cart, ?int $bid = BID): void
    {
        $session = Session::handle();
        $session->set($this->getSessionKey($bid), $cart);
        $session->save();
        $this->cache->forget($this->getCacheKey($bid));
    }

    /**
     * カートの探索
     *
     * @param int $bid
     *
     * @return Cart
     */
    public function find(int $bid = BID): Cart
    {
        $cacheKey = $this->getCacheKey($bid);
        if ($this->cache->has($cacheKey)) {
            return $this->cache->get($cacheKey);
        }

        $session = Session::handle();
        $sessionCart = $session->get(self::SESSION_PREFIX . $bid);

        if ($sessionCart === false) {
            return new Cart();
        }

        $cart = clone $sessionCart;

        $eids = array_map(
            function (Item $item) {
                return $item->getEntry()->getId();
            },
            $cart->getItems()
        );

        $this->mergeEntry($cart, $eids);

        $this->cache->put($cacheKey, $cart);

        return $cart;
    }

    /**
     * 保存されているカート内容を削除する
     *
     * @param int $bid
     */
    public function delete(int $bid = BID): void
    {
        $session = Session::handle();
        $session->delete($this->getSessionKey($bid));
        $session->save();
        $this->cache->forget($this->getCacheKey($bid));
    }

    /**
     * エントリー変数をカートアイテムにマージする
     *
     * @param Cart $cart
     * @param int[] $eids
     */
    private function mergeEntry(Cart $cart, array $eids): void
    {
        $entries = $this->EntryRepository->findByIds($eids);
        foreach ($cart->getItems() as $item) {
            $entryAry = array_filter(
                $entries,
                function (Entry $Entry) use ($item) {
                    return $Entry->getid() === $item->getEntry()->getId();
                }
            );

            $entry = array_values($entryAry)[0];

            $item->setEntry($entry);
        }
    }

    /**
     * カートアイテムが存在するかどうか
     *
     * @param int $bid
     *
     * @return bool
     */
    public function isEmpty(int $bid = BID): bool
    {
        $session = Session::handle();
        $Cart = $session->get(self::SESSION_PREFIX . $bid);

        return !$Cart || count($Cart->getItems()) === 0;
    }

    /**
     * カート内容を保存するキャッシュキーを取得する
     *
     * @param int $bid
     */
    private function getCacheKey(int $bid): string
    {
        return self::CACHE_PREFIX . $bid;
    }

    /**
     * カート内容を保存するセッションキーを取得する
     *
     * @param int $bid
     */
    private function getSessionKey(int $bid): string
    {
        return self::SESSION_PREFIX . $bid;
    }
}
