<?php

namespace Acms\Plugins\ShoppingCart\Repositories;

use Acms\Plugins\ShoppingCart\Entities\Master\ShippingGroupCalcMethod;
use Acms\Plugins\ShoppingCart\Entities\Master\Status;
use Acms\Plugins\ShoppingCart\Entities\Addition;
use Acms\Plugins\ShoppingCart\Entities\Media;
use Acms\Plugins\ShoppingCart\Entities\Payment;
use Acms\Plugins\ShoppingCart\Entities\PaymentRule;
use Acms\Plugins\ShoppingCart\Entities\Shipping;
use Acms\Plugins\ShoppingCart\Entities\ShippingGroup;
use DB;
use SQL;
use Field;

/**
 * 設定のインタフェース
 */
class ConfigRepository
{
    /**
     * 通常配送の配送グループID
     *
     * @var string
     */
    public const DEFAULT_SHIPPING_GROUP_ID = 'group_default';

    /**
     * @var \Field
     */
    protected $Config;

    public function __construct(Field $Config)
    {
        $this->Config = $Config;
    }

    public function get($key, $default = null, $i = 0)
    {
        return $this->Config->get($key, $default, $i);
    }

    public function getArray($key, $strict = false)
    {
        return $this->Config->getArray($key, $strict);
    }

    /**
     * すべての決済方法を配列で取得する
     *
     * @return Payment[]
     */
    public function findAllPayments(): array
    {
        $mids = array_filter(
            $this->getArray('shopping-cart_method-img'),
            function (?string $mid) {
                return !is_null($mid);
            }
        );

        $DB = DB::singleton(dsn());
        $SQL = SQL::newSelect('media');
        $SQL->addLeftJoin('blog', 'blog_id', 'media_blog_id');
        $SQL->addWhereIn('media_id', array_unique($mids));

        $medias = [];
        foreach ($DB->query($SQL->get(dsn()), 'all') as $row) {
            $Media = new Media();
            $Media->setId(intval($row['media_id']));
            $Media->setBid(intval($row['media_blog_id']));
            $Media->setUid(intval($row['media_user_id']));
            $Media->setStatus($row['media_status']);
            $Media->setPath($row['media_path']);
            $Media->setThumbnail($row['media_thumbnail']);
            $Media->setImageSize($row['media_image_size']);
            $Media->setFileSize(intval($row['media_file_size']));
            $Media->setFileName($row['media_file_name']);
            $Media->setType($row['media_type']);
            $Media->setExtension($row['media_extension']);
            $Media->setOriginal($row['media_original']);
            $Media->setUpdateDate(new \DateTimeImmutable($row['media_update_date']));
            $Media->setUploadDate(new \DateTimeImmutable($row['media_upload_date']));
            $Media->setCaption($row['media_field_1']);
            $Media->setLink($row['media_field_2']);
            $Media->setAlt($row['media_field_3']);
            $Media->setText($row['media_field_4']);
            $Media->setPdfPage(intval($row['media_field_5']));
            $Media->setFocalPoint($row['media_field_6']);

            $medias[$row['media_id']] = $Media;
        }

        $Payments = array_map(
            function (
                ?string $block,
                ?string $status,
                ?string $name,
                ?string $commissionPaid,
                ?string $isRuleEnabled,
                ?string $description,
                ?string $mailMessage,
                ?string $mid
            ) use ($medias) {
                // ブロック名が空のデータは除外する
                if (empty($block)) {
                    return new \stdClass();
                }

                $PaymentRules = array_map(
                    function (
                        ?string $priceMin,
                        ?string $priceMax,
                        ?string $commission
                    ) {
                        $PaymentRule = new PaymentRule();
                        $PaymentRule->setPriceMin(intval($priceMin));
                        $PaymentRule->setPriceMax(is_numeric($priceMax) ? intval($priceMax) : null);
                        $PaymentRule->setCommission(intval($commission));

                        return $PaymentRule;
                    },
                    $this->getArray('shopping-cart_payment-rule-group-price-min-' . $block),
                    $this->getArray('shopping-cart_payment-rule-group-price-max-' . $block),
                    $this->getArray('shopping-cart_payment-rule-group-commission-' . $block)
                );

                $Payment = new Payment();
                $Payment->setBlock($block);
                $Payment->setStatus($status ?: Status::OPEN);
                $Payment->setName(strval($name));
                $Payment->setCommissionPaid(intval($commissionPaid));
                $Payment->setIsRuleEnabled($isRuleEnabled === 'on');
                $Payment->setDescription(strval($description));
                $Payment->setMailMessage(strval($mailMessage));
                $Payment->setMedia(array_key_exists($mid, $medias) ? $medias[$mid] : null);
                $Payment->setPaymentRules($PaymentRules);

                return $Payment;
            },
            $this->getArray('shopping-cart_method-block'),
            $this->getArray('shopping-cart_method-status'),
            $this->getArray('shopping-cart_method-name'),
            $this->getArray('shopping-cart_method-commission'),
            $this->getArray('shopping-cart_method-enable-rule'),
            $this->getArray('shopping-cart_method-description'),
            $this->getArray('shopping-cart_method-mail-message'),
            $this->getArray('shopping-cart_method-img')
        );

        return array_values(
            array_filter(
                $Payments,
                function (object $Payment) {
                    return $Payment instanceof Payment;
                }
            )
        );
    }

    /**
     * 指定した条件の決済方法を取得する
     *
     * @param string[]|null $blocks
     * @param string|null $status
     *
     * @return Payment[]
     */
    public function findPayments(?array $blocks = null, ?string $status = null): array
    {
        return array_values(
            array_filter(
                $this->findAllPayments(),
                function (Payment $Payment) use ($blocks, $status) {
                    return (
                        (is_null($blocks) || in_array($Payment->getBlock(), $blocks, true)) &&
                        (is_null($status) || $Payment->getStatus() === $status)
                    );
                }
            )
        );
    }

    /**
     * 指定した条件のの決済方法を取得する
     * @param string|null $block
     * @param string|null $status
     *
     * @return Payment|null
     */
    public function findPayment(?string $block = null, ?string $status = null): ?Payment
    {
        foreach ($this->findAllPayments() as $Payment) {
            if (
                1 &&
                (is_null($block) || $Payment->getBlock() === $block) &&
                (is_null($status) || $Payment->getStatus() === $status)
            ) {
                return $Payment;
            }
        }

        return null;
    }

    /**
     * すべての配送グループを配列で取得する
     *
     * @return array
     */
    public function findAllShippingGroups(): array
    {
        $ids = array_merge(
            [self::DEFAULT_SHIPPING_GROUP_ID],
            $this->getArray('shopping-cart_shipping-group-id')
        );

        $statuses = array_merge(
            [Status::OPEN],
            $this->getArray('shopping-cart_shipping-group-status')
        );

        $names = array_merge(
            [$this->get('shopping-cart_shipping-group-default-name', '')],
            $this->getArray('shopping-cart_shipping-group-name')
        );

        $groupFees = array_merge(
            [$this->get('shopping-cart_shipping-group-default-fee')],
            $this->getArray('shopping-cart_shipping-group-price')
        );

        $calcMethods = array_merge(
            [$this->get('shopping-cart_shipping-group-default-calc-method')],
            $this->getArray('shopping-cart_shipping-group-calc-method')
        );

        $ShippingGroups = array_map(
            function (
                ?string $id,
                ?string $status,
                ?string $name,
                ?string $groupFee,
                ?string $calcMethod
            ) {
                // IDが空のデータは除外する
                if (empty($id)) {
                    return new \stdClass();
                }

                $Shippings = array_map(
                    function (
                        ?string $shippingName,
                        ?string $shippingFee
                    ) {
                        $Shipping = new Shipping();
                        $Shipping->setName(strval($shippingName));
                        $Shipping->setFee(is_numeric($shippingFee) ? intval($shippingFee) : null);

                        return $Shipping;
                    },
                    $this->getArray('shopping-cart_region-name-' . $id),
                    $this->getArray('shopping-cart_region-fee-' . $id)
                );

                $ShippingGroup = new ShippingGroup();
                $ShippingGroup->setId($id);
                $ShippingGroup->setStatus($status ?: Status::OPEN);
                $ShippingGroup->setName(strval($name));
                $ShippingGroup->setGroupFee(intval($groupFee));
                $ShippingGroup->setCalcMethod($calcMethod ?: ShippingGroupCalcMethod::ITEM);
                $ShippingGroup->setShippings($Shippings);
                return $ShippingGroup;
            },
            $ids,
            $statuses,
            $names,
            $groupFees,
            $calcMethods
        );

        return array_values(
            array_filter(
                $ShippingGroups,
                function (object $ShippingGroup) {
                    return $ShippingGroup instanceof ShippingGroup;
                }
            )
        );
    }

    /**
     * 指定した条件の配送グループを取得する
     *
     * @param string|null $id
     * @param string|null $status
     *
     * @return ShippingGroup|null
     */
    public function findShippingGroup(?string $id = null, ?string $status = null): ?ShippingGroup
    {
        foreach ($this->findAllShippingGroups() as $ShippingGroup) {
            /** @var ShippingGroup $ShippingGroup */
            if (
                1 &&
                (is_null($id) || $ShippingGroup->getId() === $id) &&
                (is_null($status) || $ShippingGroup->getStatus() === $status)
            ) {
                return $ShippingGroup;
            }
        }

        return null;
    }

    /**
     * 指定した配送グループIDの配送グループを配列で取得する
     *
     * @param string[]|null $ids
     * @param string|null $status
     *
     * @return ShippingGroup[]
     */
    public function findShippingGroups(?array $ids = null, ?string $status = null): array
    {
        return array_values(
            array_filter(
                $this->findAllShippingGroups(),
                function (ShippingGroup $ShippingGroup) use ($ids, $status) {
                    return (
                        (is_null($ids) || in_array($ShippingGroup->getId(), $ids, true)) &&
                        (is_null($status) || $ShippingGroup->getStatus() === $status)
                    );
                }
            )
        );
    }

    /**
     * 指定した追加手数料IDの追加手数料を配列で取得する
     *
     * @param array|null $ids 追加手数料ID
     * @param string|null $status ステータス
     *
     * @return Addition[]
     */
    public function findAdditions(?array $ids = null, ?string $status = null): array
    {
        return array_values(
            array_filter(
                $this->findAllAdditions(),
                function (Addition $Addition) use ($ids, $status) {
                    return (
                        (is_null($ids) || in_array($Addition->getId(), $ids, true)) &&
                        (is_null($status) || $Addition->getStatus() === $status)
                    );
                }
            )
        );
    }

    /**
     * すべての追加手数料を配列で取得する
     *
     * @return Addition[]
     */
    public function findAllAdditions(): array
    {
        $Additions = array_map(
            function (?string $id, ?string $name, ?string $fee, ?string $status) {
                // IDが空のデータは除外する
                if (empty($id)) {
                    return new \stdClass();
                }

                $Addition = new Addition();
                $Addition->setId($id);
                $Addition->setName(strval($name));
                $Addition->setFee(intval($fee));
                $Addition->setStatus($status ?: Status::OPEN);

                return $Addition;
            },
            $this->getArray('shopping-cart_addition-id'),
            $this->getArray('shopping-cart_addition-name'),
            $this->getArray('shopping-cart_addition-fee'),
            $this->getArray('shopping-cart_addition-status')
        );

        return array_values(
            array_filter(
                $Additions,
                function (object $Addition) {
                    return $Addition instanceof Addition;
                }
            )
        );
    }
}
