<?php

namespace Acms\Plugins\ShoppingCart\Repositories;

use DateTimeImmutable;
use Acms\Plugins\ShoppingCart\Entities\Entry;
use Acms\Plugins\ShoppingCart\Entities\ShippingGroup;
use Acms\Plugins\ShoppingCart\Repositories\ConfigRepository;
use App;
use SQL;
use DB;

/**
 * エントリーのリポジトリ
 */
class EntryRepository
{
    /**
     * @var ConfigRepository
     */
    protected ConfigRepository $Config;

    public function __construct()
    {
        $this->Config = App::make('shopping-cart.repository.config');
    }

    /**
     * 指定したエントリーIDのエントリーを取得する
     * @param int $eid
     *
     * @return Entry
     */
    public function findById(int $eid): Entry
    {
        $DB = DB::singleton(dsn());
        $entry = $DB->query($this->buildQuery([$eid]), 'row');
        $EntryField = loadEntryField($eid);
        $ShippingGroup = $this->Config->findShippingGroup(
            $EntryField->get(
                'item_shipping_group_id',
                $this->Config::DEFAULT_SHIPPING_GROUP_ID
            )
        );

        return $this->createEntry(
            $entry,
            $EntryField,
            [$ShippingGroup->getId() => $ShippingGroup]
        );
    }

    /**
     * 指定したエントリーIDのエントリーを取得する
     * @param int[] $eids
     *
     * @return Entry[]
     */
    public function findByIds(array $eids): array
    {
        $DB = DB::singleton(dsn());

        // エントリーフィールド
        $EntryFields = eagerLoadField($eids, 'eid');

        $shippingGroupIds = array_map(
            function (\Field $Field) {
                return $Field->get(
                    'item_shipping_group_id',
                    $this->Config::DEFAULT_SHIPPING_GROUP_ID
                );
            },
            array_values($EntryFields)
        );

        $ShippingGroups = array_reduce(
            $this->Config->findShippingGroups($shippingGroupIds),
            function (array $array, ShippingGroup $ShippingGroup) {
                return $array + [$ShippingGroup->getId() => $ShippingGroup];
            },
            []
        );

        return array_map(
            function (array $entry) use ($EntryFields, $ShippingGroups) {
                return $this->createEntry(
                    $entry,
                    $EntryFields[intval($entry['entry_id'])],
                    $ShippingGroups
                );
            },
            $DB->query($this->buildQuery($eids), 'all')
        );
    }

    /**
     * sqlの組み立て
     *
     * @param int[] $eids
     *
     * @return string
     */
    private function buildQuery(array $eids): string
    {
        $SQL = SQL::newSelect('entry');
        $SQL->addSelect('entry_id');
        $SQL->addSelect('entry_code');
        $SQL->addSelect('entry_status');
        $SQL->addSelect('entry_approval');
        $SQL->addSelect('entry_form_status');
        $SQL->addSelect('entry_sort');
        $SQL->addSelect('entry_user_sort');
        $SQL->addSelect('entry_category_sort');
        $SQL->addSelect('entry_title');
        $SQL->addSelect('entry_link');
        $SQL->addSelect('entry_datetime');
        $SQL->addSelect('entry_start_datetime');
        $SQL->addSelect('entry_end_datetime');
        $SQL->addSelect('entry_posted_datetime');
        $SQL->addSelect('entry_updated_datetime');
        $SQL->addSelect('entry_summary_range');
        $SQL->addSelect('entry_indexing');
        $SQL->addSelect('entry_primary_image');
        $SQL->addSelect('entry_current_rev_id');
        $SQL->addSelect('entry_last_update_user_id');
        $SQL->addSelect('entry_hash');
        $SQL->addSelect('entry_category_id');
        $SQL->addSelect('entry_user_id');
        $SQL->addSelect('entry_form_id');
        $SQL->addSelect('entry_blog_id');
        $SQL->addSelect('entry_delete_uid');

        $SQL->addWhereIn('entry_id', $eids);

        $SQL->setOrder('entry_id');

        return $SQL->get(dsn());
    }

    /**
     * Entryオブジェクトの組み立て
     *
     * @param array $entry
     * @param \Field $Field
     * @param ShippingGroup[] $ShippingGroups
     *
     * @return Entry
     */
    private function createEntry(array $entry, \Field $Field, array $ShippingGroups): Entry
    {
        $Entry = new Entry();
        $Entry->setId(intval($entry['entry_id']));
        $Entry->setCode($entry['entry_code']);
        $Entry->setStatus($entry['entry_status']);
        $Entry->setApproval($entry['entry_approval']);
        $Entry->setFormStatus($entry['entry_form_status']);
        $Entry->setSort(intval($entry['entry_sort']));
        $Entry->setUserSort(intval($entry['entry_user_sort']));
        $Entry->setCategorySort(intval($entry['entry_category_sort']));
        $Entry->setTitle($entry['entry_title']);
        $Entry->setLink($entry['entry_link']);
        $Entry->setDatetime(new DateTimeImmutable($entry['entry_datetime']));
        $Entry->setStartDatetime(new DateTimeImmutable($entry['entry_start_datetime']));
        $Entry->setEndDatetime(new DateTimeImmutable($entry['entry_end_datetime']));
        $Entry->setPostedDatetime(new DateTimeImmutable($entry['entry_posted_datetime']));
        $Entry->setUpdatedDatetime(new DateTimeImmutable($entry['entry_updated_datetime']));
        $Entry->setSummaryRange(
            !is_null($entry['entry_summary_range']) ? intval($entry['entry_summary_range']) : null
        );
        $Entry->setIndexing($entry['entry_indexing']);
        $Entry->setPrimaryImage(
            !is_null($entry['entry_primary_image']) ? intval($entry['entry_primary_image']) : null
        );
        $Entry->setCurrentRevId(intval($entry['entry_current_rev_id']));
        $Entry->setLastUpdateUserId(intval($entry['entry_last_update_user_id']));
        $Entry->setHash($entry['entry_hash']);
        $Entry->setCategoryId(
            !is_null($entry['entry_category_id']) ? intval($entry['entry_category_id']) : null
        );
        $Entry->setUserId(intval($entry['entry_user_id']));
        $Entry->setFormId(intval($entry['entry_form_id']));
        $Entry->setBlogId(intval($entry['entry_blog_id']));
        $Entry->setDeleteUid(
            !is_null($entry['entry_delete_uid']) ? intval($entry['entry_delete_uid']) : null
        );
        $Entry->setField($Field);
        $Entry->setShippingGroup($ShippingGroups[$Field->get(
            'item_shipping_group_id',
            $this->Config::DEFAULT_SHIPPING_GROUP_ID
        )]);

        return $Entry;
    }
}
