<?php

namespace Acms\Plugins\ShoppingCart;

use Acms\Services\Common\HookFactory;
use Acms\Services\Common\CorrectorFactory;
use Acms\Services\Common\InjectTemplate;
use App;
use ACMS_App;
use Storage;
use Config;
use Acms\Plugins\ShoppingCart\Repositories\CartRepository;
use Acms\Plugins\ShoppingCart\Repositories\ConfigRepository;
use Acms\Plugins\ShoppingCart\Repositories\EntryRepository;
use Acms\Plugins\ShoppingCart\Services\Encrypter;
use Acms\Plugins\ShoppingCart\Services\OrderService;
use Acms\Plugins\ShoppingCart\Services\ShoppingService;
use Acms\Plugins\ShoppingCart\Services\TaxService;
use Acms\Plugins\ShoppingCart\Services\Payment\PaymentMethodFactory;
use Acms\Plugins\ShoppingCart\Services\Payment\Methods\Cash;
use Acms\Plugins\ShoppingCart\Services\Payment\Methods\CreditCardFactory;
use Acms\Plugins\ShoppingCart\Services\Payment\Methods\CreditCardAdapter;
use Acms\Plugins\ShoppingCart\Plugins\Square\Services\OAuth\SquareOAuthFactory;
use Acms\Plugins\ShoppingCart\Plugins\Square\Services\OAuth\SquareOAuthService;
use Acms\Plugins\ShoppingCart\Plugins\Square\Services\OAuth\SquareSandboxOAuthService;
use Acms\Plugins\ShoppingCart\Plugins\Square\Services\SquarePaymentService;
use Acms\Plugins\ShoppingCart\Plugins\Square\Services\Methods\SquareCreditCard;

class ServiceProvider extends ACMS_App
{
    /**
     * @var string
     */
    public $version = '2.2.3';

    /**
     * @var string
     */
    public $name = 'ShoppingCart';

    /**
     * @var string
     */
    public $author = 'com.appleple';

    /**
     * @var bool
     */
    public $module = false;

    /**
     * @var false|string
     */
    public $menu = "shopping_cart_index";

    /**
     * @var string
     */
    public $desc = 'a-blog cmsでECサイトを制作するための拡張アプリです。';

    /**
     * @var array
     */
    private $themes = [
        'ec',
        'square@ec'
    ];

    /**
     * サービスの初期処理
     */
    public function init()
    {
        require_once dirname(__FILE__) . '/vendor/autoload.php';

        $Config = Config::loadDefaultField();
        $Config->overload(Config::loadBlogConfig(BID));

        /**
         * Register Shopping Cart Repository
         */
        App::singleton('shopping-cart.repository.cart', CartRepository::class);

        App::singleton('shopping-cart.repository.config', function () use ($Config) {
            return new ConfigRepository($Config);
        });

        App::singleton('shopping-cart.repository.entry', EntryRepository::class);

        /**
         * Register Shopping Cart Service
         */
        App::singleton('shopping-cart.service.order', OrderService::class);

        App::singleton('shopping-cart.service.shopping', ShoppingService::class);

        App::singleton('shopping-cart.service.tax', TaxService::class);

        App::singleton('shopping-cart.service.encrypter', function () {
            return new Encrypter('aes-128-gcm');
        });

        /**
         * Register Payment Method
         */
        $PaymentMethod = PaymentMethodFactory::singleton();
        $PaymentMethod->attachWithDefault('cash', Cash::class);
        $PaymentMethod->attach('creditcard', CreditCardAdapter::class);

        /**
         * Register Square Payment Method
         */
        $CreditCard = CreditCardFactory::singleton();
        $CreditCard->attach('square', SquareCreditCard::class);

        /**
         * Register Square OAuth Service
         */
        App::singleton('square.service.oauth', SquareOAuthFactory::class);
        App::singleton('square.service.oauth.production', function () use ($Config) {
            return new SquareOAuthService(
                $Config->get('SquareApplicationID'),
                $Config->get('SquareApplicationSecret')
            );
        });
        App::singleton('square.service.oauth.sandbox', function () use ($Config) {
            return new SquareSandboxOAuthService(
                $Config->get('SandboxSquareApplicationID'),
                $Config->get('SandboxSquareApplicationSecret')
            );
        });

        /**
         * Register Square Payment Service
         */
        App::singleton('square.service.payment', SquarePaymentService::class);

        /**
         * Attach Hook
         */
        $hook = HookFactory::singleton();
        $hook->attach('shopping-cart.hook', new Hook());

        /**
         * Attach Coorrector
         */
        $corrector = CorrectorFactory::singleton();
        $corrector->attach('shopping-cart.corrector', new Corrector());

        /**
         * Inject Template
         */
        $inject = InjectTemplate::singleton();
        $inject->add(
            'admin-module-config-Order_Summary',
            PLUGIN_DIR . $this->name . '/template/admin/config/order/summary_body.html'
        );
        $inject->add(
            'admin-module-config-ShoppingCart_Settings',
            PLUGIN_DIR . $this->name . '/template/admin/config/shopping-cart/settings_body.html'
        );
        $inject->add(
            'admin-module-config-ShoppingCart_Summary',
            PLUGIN_DIR . $this->name . '/template/admin/config/shopping-cart/summary_body.html'
        );
        $inject->add(
            'admin-module-select',
            PLUGIN_DIR . $this->name . '/template/admin/module/select.html'
        );
        $inject->add('square-js', PLUGIN_DIR . $this->name . '/Plugins/Square/template/js.html');
        if (ADMIN === 'app_' . $this->menu) {
            $inject->add(
                'admin-topicpath',
                PLUGIN_DIR . $this->name . '/template/admin/topicpath.html'
            );
            $inject->add('admin-main', PLUGIN_DIR . $this->name . '/template/admin/main.html');
        } elseif (ADMIN === 'app_shopping_cart_square_callback') {
            $inject->add(
                'admin-main',
                PLUGIN_DIR . $this->name . '/Plugins/Square/template/callback.html'
            );
        }

        /**
         * Square Webhook
         */
        if (strpos(HTTP_REQUEST_URL, 'webhook/shopping-cart/square')) {
            require_once dirname(__FILE__) . '/Plugins/Square/Webhook/Router.php';
        }
    }

    /**
     * インストールする前の環境チェック処理
     *
     * @return bool
     */
    public function checkRequirements()
    {
        return true;
    }

    /**
     * インストールするときの処理
     * データベーステーブルの初期化など
     *
     * @return void
     */
    public function install()
    {
        // テーマの移動
        foreach ($this->themes as $theme) {
            if (!Storage::exists(SCRIPT_DIR . THEMES_DIR . $theme)) {
                // テーマが存在しなければコピー
                Storage::copyDirectory(
                    PLUGIN_LIB_DIR . $this->name . '/themes/' . $theme,
                    SCRIPT_DIR . THEMES_DIR . $theme
                );
            }
        }

        $this->putConfig();
    }

    /**
     * アンインストールするときの処理
     * データベーステーブルの始末など
     *
     * @return void
     */
    public function uninstall()
    {
    }

    /**
     * アップデートするときの処理
     *
     * @return bool
     */
    public function update()
    {
        return true;
    }

    /**
     * 有効化するときの処理
     *
     * @return bool
     */
    public function activate()
    {
        return true;
    }

    /**
     * 無効化するときの処理
     *
     * @return bool
     */
    public function deactivate()
    {
        return true;
    }

    /**
     * 拡張アプリのデフォルト設定をconfig.system.yamlに追記
     *
     * @return void
     */
    private function putConfig()
    {
        $config = Storage::get(CONFIG_FILE);
        $shoppingCartConfig = Storage::get(PLUGIN_LIB_DIR . $this->name . '/config.system.yaml');
        if (preg_match('/(#BEGIN_ShoppingCartConfig)[\s\S]*(#END_ShoppingCartConfig)/', $config)) {
            Storage::put(
                CONFIG_FILE,
                preg_replace(
                    '/(#BEGIN_ShoppingCartConfig)[\s\S]*(#END_ShoppingCartConfig)/',
                    $shoppingCartConfig,
                    $config
                )
            );
        } else {
            Storage::put(CONFIG_FILE, $config . $shoppingCartConfig);
        }
    }
}
