<?php

namespace Acms\Plugins\ShoppingCart\Services;

use Acms\Plugins\ShoppingCart\Entities\Addition;
use Acms\Plugins\ShoppingCart\Entities\Order;
use Acms\Plugins\ShoppingCart\Entities\OrderItem;
use Acms\Plugins\ShoppingCart\Entities\PaymentRule;
use Acms\Plugins\ShoppingCart\Entities\TaxAmount;
use Acms\Plugins\ShoppingCart\Entities\Shipping;
use Acms\Plugins\ShoppingCart\Entities\ShippingGroup;
use Acms\Plugins\ShoppingCart\Entities\Master\ShippingGroupCalcMethod;
use Acms\Plugins\ShoppingCart\Repositories\ConfigRepository;
use Acms\Plugins\ShoppingCart\Services\TaxService;
use App;
use RuntimeException;

class OrderService
{
    /**
     * @var TaxService
     */
    protected $TaxService;

    /**
     * @var ConfigRepository
     */
    protected $Config;

    /**
     * constractor
     */
    public function __construct()
    {
        $this->Config = App::make('shopping-cart.repository.config');
        $this->TaxService = App::make('shopping-cart.service.tax');
    }

    /**
     * 小計金額を取得する
     *
     * @param Order $Order
     *
     * @return int 小計金額
     */
    public function getSubtotal(Order $Order): int
    {
        return array_reduce(
            $Order->getTaxAmounts(),
            function ($subtotal, TaxAmount $TaxAmount) {
                return $subtotal + $TaxAmount->getExcludingTax() + $TaxAmount->getTax();
            },
            0
        );
    }

    /**
     * 配送料を取得する
     *
     * @param Order $Order
     *
     * @return int 配送料
     */
    public function getShippingFee(Order $Order): int
    {
        $calcMethod = $this->Config->get('shopping-cart_calc-method', '');
        if ($calcMethod === 'sum') {
            return $this->getShippingFeeAmount($Order);
        } elseif ($calcMethod === 'max') {
            return $this->getHighestShippingFee($Order);
        }

        return 0;
    }

    /**
     * 決済手数料を取得する
     *
     * @param Order $Order
     *
     * @return int 決済手数料
     */
    public function getCommissionPaid(Order $Order): int
    {
        if (is_null($Order->getPayment())) {
            return 0;
        }

        $Payment = $Order->getPayment();
        $commissionPaid = $Payment->getCommissionPaid();

        if (!$Payment->getIsRuleEnabled()) {
            return $commissionPaid;
        }

        $base = $Order->getSubtotal() + $Order->getShippingFee() + $Order->getAdditionalFee();
        foreach ($Payment->getPaymentRules() as $PaymentRule) {
            /** @var PaymentRule $PaymentRule */
            $thresholdPrice = $base + $PaymentRule->getCommission();
            if (
                1 &&
                $PaymentRule->getPriceMin() <= ($thresholdPrice) &&
                (
                    is_null($PaymentRule->getPriceMax()) ||
                    $PaymentRule->getPriceMax() >= ($thresholdPrice)
                )
            ) {
                $commissionPaid = $PaymentRule->getCommission();
                break;
            }
        }

        return $commissionPaid;
    }

    /**
     * 追加手数料を取得する
     *
     * @param Order $Order
     *
     * @return int 追加手数料
     */
    public function getAdditionalFee(Order $Order): int
    {
        return array_reduce(
            $Order->getAdditions(),
            function (int $additionalFee, Addition $Addtion) {
                return $additionalFee + $Addtion->getFee();
            },
            0
        );
    }

    /**
     * 合計金額を計算する
     *
     * @param Order $Order
     *
     * @return int 合計金額
     */
    public function getTotal(Order $Order): int
    {
        $subtotal = $Order->getSubtotal();
        $shippingFee = $Order->getShippingFee();
        $commissionPaid = $Order->getCommissionPaid();
        $additionalFee = $Order->getAdditionalFee();
        return $subtotal + $shippingFee + $commissionPaid + $additionalFee;
    }

    /**
     * 配送グループで設定された送料を取得する
     *
     * @param ShippingGroup $ShippingGroup 配送グループ
     * @param Order $Order 注文情報
     * @param OrderItem|null $OrderItem 注文商品情報
     *
     * @return int 送料
     */
    public function getShippingGroupFee(ShippingGroup $ShippingGroup, Order $Order, ?OrderItem $OrderItem = null): int
    {
        $shippingFee = $ShippingGroup->getGroupFee();

        if (!is_null($Order->getDeliveryArea())) {
            // 地域ごとの送料設定が存在している場合は、地域の送料を適用する
            foreach ($ShippingGroup->getShippings() as $Shipping) {
                /** @var Shipping $Shipping */
                if ($Shipping->getName() === $Order->getDeliveryArea() && !is_null($Shipping->getFee())) {
                    $shippingFee = $Shipping->getFee();
                    break;
                }
            }
        }

        if ($ShippingGroup->getCalcMethod() === ShippingGroupCalcMethod::ITEM) {
            return $shippingFee;
        } elseif ($ShippingGroup->getCalcMethod() === ShippingGroupCalcMethod::QUANTITY && !is_null($OrderItem)) {
            return $shippingFee * $OrderItem->getQuantity();
        }

        return $shippingFee;
    }

    /**
     * 最も高い送料を取得する
     *
     * @param Order $Order
     *
     * @return int 送料
     */
    private function getHighestShippingFee(Order $Order): int
    {
        $shippingFees = array_map(
            function (OrderItem $OrderItem) use ($Order) {
                return $this->getShippingGroupFee($OrderItem->getShippingGroup(), $Order, $OrderItem);
            },
            $Order->getOrderItems()
        );

        if (count($shippingFees) < 1) {
            return 0;
        }

        return max($shippingFees);
    }

    /**
     * 合計した送料を取得する
     *
     * @param Order $Order
     *
     * @return int 送料
     */
    private function getShippingFeeAmount(Order $Order): int
    {
        return array_reduce(
            $Order->getOrderItems(),
            function (int $amount, OrderItem $OrderItem) use ($Order) {
                return $amount + $this->getShippingGroupFee($OrderItem->getShippingGroup(), $Order, $OrderItem);
            },
            0
        );
    }
}
