<?php

namespace Acms\Plugins\ShoppingCart\Services;

use Acms\Plugins\ShoppingCart\Entities\PaymentRule;
use Acms\Plugins\ShoppingCart\Entities\Item;
use Acms\Plugins\ShoppingCart\Entities\Order;
use Acms\Plugins\ShoppingCart\Entities\OrderItem;
use Acms\Plugins\ShoppingCart\Entities\Payment;
use Acms\Plugins\ShoppingCart\Exceptions\ShoppingException;
use Acms\Plugins\ShoppingCart\Repositories\ConfigRepository;
use Acms\Plugins\ShoppingCart\Services\Payment\PaymentResult;
use Acms\Plugins\ShoppingCart\Services\Payment\PaymentMethodFactory;
use Acms\Plugins\ShoppingCart\Services\Payment\PaymentMethodInterface;
use Acms\Plugins\ShoppingCart\Services\OrderService;
use Acms\Plugins\ShoppingCart\Services\TaxService;
use App;
use Field;
use Common;
use SQL;
use Session;

class ShoppingService
{
    /**
     * カート内容を保存する使用するセッションキーのプレフィックス
     *
     * @var string
     */
    private const SESSION_PROCESSING_ORDER_KEY = 'processingOrder';

    /**
     * @var OrderService
     */
    protected $OrderService;

    /**
     * @var TaxService
     */
    protected $TaxService;

    /**
     * @var ConfigRepository
     */
    protected $Config;

    /**
     * constractor
     */
    public function __construct()
    {
        $this->OrderService = App::make('shopping-cart.service.order');
        $this->TaxService = App::make('shopping-cart.service.tax');
        $this->Config = App::make('shopping-cart.repository.config');
    }

    /**
     * 注文情報のバリデート
     *
     * @param Order $Order
     *
     * @return bool
     */
    public function validate(Order $Order): bool
    {
        $OrderItems = $Order->getOrderItems();

        if (count($OrderItems) <= 0) {
            // 注文商品が存在しないためエラー
            return false;
        }

        foreach ($OrderItems as $orderItem) {
            /** @var OrderItem $orderItem */
            // 在庫が無制限かチェックし、制限ありなら在庫数をチェック
            if ($orderItem->isStockManagementEnabled()) {
                // 在庫管理が有効

                if (intval($orderItem->getEntry()->getField()->get('item_stock')) < 1) {
                    return false;
                }

                if ($orderItem->isStockShortage()) {
                    return false;
                }
            }
        }

        return true;
    }

    /**
     * 在庫情報のチェックとロック
     *
     * @param Order $Order
     * @param \Acms\Services\Database\Engine\Base $DB トランザクション制御されているDatabase
     *
     * @return bool
     */
    public function checkStock(Order $Order, \Acms\Services\Database\Engine\Base $DB): bool
    {
        // 在庫制限ありの商品
        $stockManagementItems = $Order->getStockManagementItems();

        $SQL = SQL::newSelect('field');
        $SQL->addSelect('field_eid');
        $SQL->addSelect('field_value');
        $SQL->addWhereOpr('field_key', 'item_stock');
        $SQL->addWhereIn(
            'field_eid',
            array_map(
                function (OrderItem $stockManagementItem) {
                    return $stockManagementItem->getEntry()->getId();
                },
                $stockManagementItems
            )
        );

        // 在庫に対してロック(select ... for update)を実行
        $rows = $DB->query($SQL->get(dsn()) . ' FOR UPDATE', 'all');

        foreach ($stockManagementItems as $stockManagementItem) {
            $stockAry = array_filter(
                $rows,
                function (array $row) use ($stockManagementItem) {
                    return intval($row['field_eid']) === $stockManagementItem->getEntry()->getId();
                }
            );

            $stock = array_values($stockAry)[0];

            if ($stock < 1) {
                return false;
            }

            if ($stockManagementItem->getQuantity() > $stock) {
                return false;
            }
        }

        return true;
    }

    /**
     * 注文情報を作成する
     *
     * @param Item[] $Items
     * @param Field|null $Field
     *
     * @return Order
     */
    public function createOrder(array $Items, ?Field $Field = null): Order
    {
        $Order = new Order();
        $Order->setProcessingId(uniqid());

        $this->setOrderItems($Order, $Items);

        $this->setTaxAmounts($Order);

        $this->setSubtotal($Order);

        $this->setServiceCharge($Order, $Field);

        // 合計金額をセット
        $this->setTotal($Order);

        return $Order;
    }

    /**
     * 注文情報を更新する
     *
     * @param Order $Order
     * @param Item[] $Items
     * @param Field $Field
     *
     * @return Order
     */
    public function updateOrder(Order $Order, array $Items, Field $Field): Order
    {

        $this->setOrderItems($Order, $Items);

        $this->setTaxAmounts($Order);

        $this->setSubtotal($Order);

        $this->setServiceCharge($Order, $Field);

        // 合計金額をセット
        $this->setTotal($Order);

        return $Order;
    }

    /**
     * 購入処理中の注文情報を取得する
     * @param string $processingId
     *
     * @return Order|null
     */
    public function loadProcessingOrder(string $processingId): ?Order
    {
        $Session = Session::handle();
        $Orders = $Session->get(self::SESSION_PROCESSING_ORDER_KEY) ?: [];

        foreach ($Orders as $Order) {
            if ($Order->getProcessingId() === $processingId) {
                return clone $Order;
            }
        }

        return null;
    }

    /**
     * 購入処理中の注文情報を保存する
     *
     * @param Order|null $Order
     *
     * @return void
     */
    public function saveProcessingOrder(?Order $Order): void
    {
        $Session = Session::handle();
        $Orders = array_filter(
            $Session->get(self::SESSION_PROCESSING_ORDER_KEY) ?: [],
            function (Order $ProcessingOrder) use ($Order) {
                return $Order->getProcessingId() !== $ProcessingOrder->getProcessingId();
            }
        );
        $Session->set(self::SESSION_PROCESSING_ORDER_KEY, array_merge($Orders, [$Order]));
        $Session->save();
    }

    /**
     * 購入処理中の注文情報を削除する
     *
     * @param Order $Order
     *
     * @return void
     */
    public function deleteProcessingOrder(Order $Order): void
    {
        $Session = Session::handle();
        $Orders = $Session->get(self::SESSION_PROCESSING_ORDER_KEY) ?: [];

        $Orders = array_filter(
            $Orders,
            function (Order $ProcessingOrder) use ($Order) {
                return $Order->getProcessingId() !== $ProcessingOrder->getProcessingId();
            }
        );

        // 配列のインデックスを初期化するためにarray_valuesを使用
        $Session->set(self::SESSION_PROCESSING_ORDER_KEY, array_values($Orders));
        $Session->save();
    }

    /**
     * 各種手数料をセットする
     *
     * @param Order $Order
     *
     * @return void
     */
    private function setTaxAmounts(Order $Order)
    {
        $TaxAmounts = $this->TaxService->createTaxAmounts($Order->getOrderItems());
        $Order->setTaxAmounts($TaxAmounts);
    }

    /**
     * 小計をセットする
     *
     * @param Order $Order
     *
     * @return void
     */
    private function setSubtotal(Order $Order)
    {
        $Order->setSubtotal($this->OrderService->getSubtotal($Order));
    }

    /**
     * 税金情報をセットする
     *
     * @param Order $Order
     * @param Item[] $Items
     *
     * @return void
     */
    private function setOrderItems(Order $Order, array $Items)
    {
        $OrderItems = array_map(
            function (Item $Item) {
                $OrderItem = new OrderItem();
                $OrderItem->setEntry($Item->getEntry());
                $OrderItem->setQuantity($Item->getQuantity());
                $OrderItem->setPrice(intval($Item->getEntry()->getField()->get('item_price')));
                $OrderItem->setTaxRate(floatval($Item->getEntry()->getField()->get('item_tax')));
                $OrderItem->setTax($this->TaxService->calcTax($OrderItem->getPrice(), $OrderItem->getTaxRate()));
                $OrderItem->setShippingGroup($Item->getEntry()->getShippingGroup());
                $OrderItem->setField($Item->getField());

                return $OrderItem;
            },
            $Items
        );

        $Order->setOrderItems($OrderItems);
    }

    /**
     * 各種手数料をセットする
     *
     * @param Order $Order
     * @param Field|null $Field
     *
     * @return void
     */
    public function setServiceCharge(Order $Order, ?Field $Field)
    {
        if (is_null($Field)) {
            $Field = new Field();
        }

        // 追加手数料情報をセット
        $this->setAdditions($Order, $Field);

        // 配送情報をセット
        $this->setShipping($Order, $Field);

        // 決済情報をセット
        $this->setPayment($Order, $Field);
    }

    /**
     * 購入処理を行う
     *
     * @param Order $Order
     * @param PaymentMethodInterface $PaymentMethod
     * @param \Acms\Services\Database\Engine\Base $DB トランザクション制御されているDatabase
     *
     * @return PaymentResult
     * @throws ShoppingException
     */
    public function processCheckout(
        Order $Order,
        PaymentMethodInterface $PaymentMethod,
        \Acms\Services\Database\Engine\Base $DB
    ) {
        // 合計金額の再計算（念の為）
        $this->calculatePrice($Order);

        if (!$this->checkStock($Order, $DB)) {
            throw new ShoppingException(
                'The stock number has been exceeded by the stock number which can be purchased.'
            );
        };

        // 決済処理実行
        $PaymentResult = $PaymentMethod->checkout();

        if ($PaymentResult->isSuccess()) {
            $this->releaseStock($Order, $DB);
        }

        return $PaymentResult;
    }

    /**
     * 合計金額を計算
     *
     * @param Order $Order
     * @return Order
     */
    private function calculatePrice(Order $Order)
    {

        $total = $this->OrderService->getTotal($Order);

        if ($total < 0) {
            $total = 0;
        }

        $Order->setTotal($total);

        return $Order;
    }

    /**
     * カート分の在庫を減らす
     *
     * @param Order $Order
     * @param \Acms\Services\Database\Engine\Base $DB トランザクション制御されているDatabase
     */
    public function releaseStock(Order $Order, \Acms\Services\Database\Engine\Base $DB): void
    {
        foreach ($Order->getOrderItems() as $OrderItem) {
            /** @var OrderItem $OrderItem */
            // 在庫が無制限かチェック
            if (!$OrderItem->isStockManagementEnabled()) {
                // 無制限であればループを抜ける
                continue;
            }

            // 制限ありなら在庫数を更新
            Common::deleteFieldCache('eid', $OrderItem->getEntry()->getId());

            $SQL = SQL::newUpdate('field');
            $SQL->addUpdate(
                'field_value',
                // 0以下にはしない
                max(intval($OrderItem->getEntry()->getField()->get('item_stock')) - $OrderItem->getQuantity(), 0)
            );
            $SQL->addWhereOpr('field_key', 'item_stock');
            $SQL->addWhereOpr('field_eid', $OrderItem->getEntry()->getId());
            $DB->query($SQL->get(dsn()), 'exec');
        }
    }

    /**
     * PaymentMethodをDIコンテナから取得する
     * 注文情報で決済方法が設定されていない場合は現金払いとして扱う
     *
     * @param Order $Order
     * @param Field $Post
     *
     * @return PaymentMethodInterface
     */
    public function createPaymentMethod(Order $Order, Field $Post): PaymentMethodInterface
    {
        $PaymentMethodFactory = PaymentMethodFactory::singleton();
        // 決済方法が設定されていない場合は現金払いとして扱う
        $blockName = $Order->getPayment() ? $Order->getPayment()->getBlock() : 'cash';
        $PaymentMethod = $PaymentMethodFactory->create($blockName);
        $PaymentMethod->setOrder($Order);
        $PaymentMethod->setPost($Post);

        return $PaymentMethod;
    }

    /**
     * 配送情報をセット
     *
     * @param Order $Order
     * @param Field $Field
     */
    private function setShipping(Order $Order, Field $Field)
    {
        $Order->setDeliveryArea($Field->get('deliver-area'));

        $Order->setShippingFee($this->OrderService->getShippingFee($Order));

        $this->setShippingFreeSubtotal($Order);
    }

    /**
     * 送料無料条件（小計）条件を満たしていれば送料を0に設定
     *
     * @param Order $Order
     */
    private function setShippingFreeSubtotal(Order $Order)
    {
        $shippingFreeSubtotal = $this->Config->get('shopping-cart_shipping-free-subtotal');
        // 送料無料条件（小計）が小計金額以上であれば送料無料
        if (is_numeric($shippingFreeSubtotal) && $Order->getSubtotal() >= intval($shippingFreeSubtotal)) {
            $Order->setShippingFee(0);
        }
    }

    /**
     * 決済情報のセット
     *
     * @param Order $Order
     * @param Field $Field
     *
     */
    private function setPayment(Order $Order, Field $Field)
    {
        $paymentMethod = $Field->get('payment-method');
        if (!empty($paymentMethod)) {
            $Payment = $this->Config->findPayment($paymentMethod);

            $Order->setPayment($Payment);
            $commissionPaid = $this->OrderService->getCommissionPaid($Order);
            $Order->setCommissionPaid($commissionPaid);
            // 決済方法の手数料も書き換える
            $Payment->setCommissionPaid($commissionPaid);
        }
    }

    /**
     * 利用可能な決済方法の取得
     *
     * @param Payment[] $Payments
     * @param Order $Order
     *
     * @return Payment[]
     */
    public function getAvailablePayments(array $Payments, Order $Order): array
    {
        $base = $Order->getSubtotal() + $Order->getShippingFee() + $Order->getAdditionalFee();
        return array_values(
            array_filter(
                $Payments,
                function (Payment $Payment) use ($base) {
                    if (!$Payment->getIsRuleEnabled()) {
                        return true;
                    }

                    foreach ($Payment->getPaymentRules() as $PaymentRule) {
                        /** @var PaymentRule $PaymentRule */
                        $thresholdPrice = $base + $PaymentRule->getCommission();
                        if (
                            1 &&
                            $PaymentRule->getPriceMin() <= ($thresholdPrice) &&
                            (
                                is_null($PaymentRule->getPriceMax()) ||
                                $PaymentRule->getPriceMax() >= ($thresholdPrice)
                            )
                        ) {
                            // 決済方法の手数料を書き換える
                            $Payment->setCommissionPaid($PaymentRule->getCommission());
                            return true;
                        }
                    }

                    return false;
                }
            )
        );
    }

    /**
     * 追加手数料情報のセット
     *
     * @param Order $Order
     * @param Field $Field
     *
     */
    private function setAdditions(Order $Order, Field $Field): void
    {
        $additions = $Field->getArray('addition');
        if (!empty($additions)) {
            $Additions = $this->Config->findAdditions($additions);

            $Order->setAdditions($Additions);
            $Order->setAdditionalFee($this->OrderService->getAdditionalFee($Order));
        }
    }

    /**
     * 合計金額のセット
     *
     * @param Order $Order
     *
     */
    public function setTotal(Order $Order)
    {
        $Order->setTotal($this->OrderService->getTotal($Order));
    }
}
