<?php

namespace Acms\Plugins\ShoppingCart\Services;

use Acms\Plugins\ShoppingCart\Entities\Item;
use Acms\Plugins\ShoppingCart\Entities\OrderItem;
use Acms\Plugins\ShoppingCart\Entities\TaxAmount;
use Acms\Plugins\ShoppingCart\Repositories\ConfigRepository;
use App;
use RuntimeException;

class TaxService
{
    /**
     * @var ConfigRepository
     */
    protected $Config;

    /**
     * constractor
     */
    public function __construct()
    {
        $this->Config = App::make('shopping-cart.repository.config');
    }

    /**
     * 消費税率ごとの税抜き価格及び税を計算する
     *
     * @param Item[]|OrderItem[] $Items
     *
     * @return TaxAmount[]
     * @throws RuntimeException
     */
    public function createTaxAmounts(array $Items)
    {
        $taxFormat = $this->Config->get('shopping-cart_tax-format', '');
        $taxCalcMethod = $this->Config->get('shopping-cart_tax-calc-method', '');

        $taxes = [];

        // 税の合計方法によって処理を分岐する
        switch ($taxCalcMethod) {
            case 'rebate':
                // 各商品の合計金額を税率別に計算する
                foreach ($Items as $Item) {
                    /** @var Item|OrderItem $Item */
                    $rate = $Item->getEntry()->getField()->get('item_tax');
                    $price = intval($Item->getEntry()->getField()->get('item_price'));
                    $quantity = $Item->getQuantity();

                    if (isset($taxes[$rate])) {
                        $taxes[$rate]['total'] += $price * $quantity;
                    } else {
                        // 税率のキーが存在しない場合、新規に作成する
                        $taxes[$rate] = [
                            'total' => $price * $quantity,
                            'excludingTax' => 0,
                            'tax' => 0
                        ];
                    }
                }

                switch ($taxFormat) {
                    case 'intax':
                        foreach ($taxes as $rate => &$tax) {
                            // $rateは税率を表す
                            $tax['tax'] = $this->roundTax(
                                $this->calcTaxFromTaxIncluded(
                                    intval($tax['total']),
                                    doubleval($rate)
                                )
                            );
                            $tax['excludingTax'] = $tax['total'] - $tax['tax'];
                        }
                        break;

                    case 'extax':
                        foreach ($taxes as $rate => &$tax) {
                            // $rateは税率を表す
                            $tax['tax'] = $this->roundTax($tax['total'] * doubleval($rate));
                            $tax['excludingTax'] = $this->roundTax($tax['total']);
                            $tax['total'] = $tax['tax'] + $tax['excludingTax'];
                        }
                        break;

                    default:
                        throw new RuntimeException('Tax Expression Format is not defined.');
                }
                break;

            case 'pileup':
                // 商品ごと
                // Warning: 商品ごとの計算は23年に法律違反になる]
                // see https://it.srad.jp/story/19/09/19/1445220/
                switch ($taxFormat) {
                    case 'intax':
                        // 各商品の「税抜き＋税」及び消費税の合計を税率別に計算する
                        foreach ($Items as $Item) {
                            /** @var Item|OrderItem $Item */
                            $rate = $Item->getEntry()->getField()->get('item_tax');
                            $price = intval($Item->getEntry()->getField()->get('item_price'));
                            $quantity = $Item->getQuantity();

                            if (isset($taxes[$rate])) {
                                $taxes[$rate]['total'] += $price * $quantity;
                                $taxes[$rate]['tax'] += $this->roundTax(
                                    $this->calcTaxFromTaxIncluded($price, doubleval($rate))
                                ) * $quantity;
                            } else {
                                // 税率のキーが存在しない場合、新規に作成する
                                $taxes[$rate] = [
                                    'total' => $price * $quantity,
                                    'excludingTax' => 0,
                                    // 各商品ごとの消費税の合計 = 商品の消費税 * 商品の数
                                    'tax' => $this->roundTax(
                                        $this->calcTaxFromTaxIncluded($price, doubleval($rate))
                                    ) * $quantity
                                ];
                            }
                        }

                        foreach ($taxes as $rate => &$tax) {
                            $tax['excludingTax'] = $tax['total'] - $tax['tax'];
                        }
                        break;

                    case 'extax':
                        // 各商品の「税抜き＋税」及び消費税の合計を税率別に計算する
                        foreach ($Items as $Item) {
                            /** @var Item|OrderItem $Item */
                            $rate = $Item->getEntry()->getField()->get('item_tax');
                            $price = intval($Item->getEntry()->getField()->get('item_price'));
                            $quantity = $Item->getQuantity();

                            if (isset($taxes[$rate])) {
                                $taxes[$rate]['total'] += $price * $quantity;
                                $taxes[$rate]['tax'] += $this->roundTax(
                                    $price * doubleval($rate)
                                ) * $quantity;
                            } else {
                                // 税率のキーが存在しない場合、新規に作成する
                                $taxes[$rate] = [
                                    'total' => $price * $quantity,
                                    'excludingTax' => 0,
                                    'tax' => $this->roundTax(
                                        $price * doubleval($rate)
                                    ) * $quantity // 各商品ごとの消費税の合計 = 商品の消費税 * 商品の数
                                ];
                            }
                        }

                        foreach ($taxes as $rate => &$tax) {
                            $tax['excludingTax'] = $tax['total'];
                            $tax['total'] = $tax['excludingTax'] + $tax['tax'];
                        }
                        break;

                    default:
                        throw new RuntimeException('Tax Expression Format is not defined.');
                }
                break;

            default:
                throw new RuntimeException('Tax calc method is not defined.');
        }

        ksort($taxes); // 税率の昇順でソート

        return array_map(
            function ($taxRate, $tax) {
                $TaxAmount = new TaxAmount();
                $TaxAmount->setTaxRate(doubleval($taxRate));
                $TaxAmount->setExcludingTax($tax['excludingTax']);
                $TaxAmount->setTax($tax['tax']);

                return $TaxAmount;
            },
            array_keys($taxes),
            $taxes
        );
    }

    /**
     * 内税価格の消費税を計算する
     *
     * @param int $taxIncluded 税込価格
     * @param double $taxRate 税率
     *
     * @return double 消費税
     */
    public function calcTaxFromTaxIncluded($taxIncluded, $taxRate)
    {
        // Warn: 税率が小数点以下2桁までであることが前提の式
        // 浮動小数点対策のためにint型にして計算する
        return intval($taxIncluded * $taxRate * 100) * (100 / intval((1 + $taxRate) * 100)) * 0.01;
    }

    /**
     * 外税価格の消費税を計算する
     *
     * @param int $taxExcluded 外税価格
     * @param double $taxRate 税率
     *
     * @return double 消費税
     */
    public function calcTaxFromTaxExcluded($taxExcluded, $taxRate)
    {
        return $taxExcluded * $taxRate;
    }

    /**
     * 金額の端数計算を行う
     *
     * @param double $amount
     *
     * @return int 端数処理済み整数値
     */
    public function roundTax($amount)
    {
        switch ($this->Config->get('shopping-cart_tax-rounding', '')) {
            case 'floor':
                return floor($amount);
            case 'round':
                return round($amount);
            case 'ceil':
                return ceil($amount);
            default:
                throw new RuntimeException('Tax rounding method is not defined.');
        }
    }

    /**
     * 税込価格を計算する
     *
     * @param  int $price 税抜き価格
     * @param  double  $taxRate 税率
     *
     * @return int 税込価格
     */
    public function calcIncludingTax($price, $taxRate)
    {
        return $price + $this->roundTax($price * $taxRate);
    }

    /**
     * 合計価格を計算する
     *
     * @param  int $price 税抜き価格
     * @param  int  $quantity 数量
     * @param  double|null  $taxRate 税率
     * @return int 合計価格
     */
    public function calcTotalPrice($price, $quantity, $taxRate)
    {
        // 税率が存在する場合
        if (!empty($taxRate)) {
            switch ($this->Config->get('shopping-cart_tax-calc-method', '')) {
                case 'rebate':
                    // 割り戻し
                    return ($price * $quantity) + $this->roundTax($price * $quantity * $taxRate);
                case 'pileup':
                    // 積み上げ
                    return $this->roundTax(
                        $price + ($price * $taxRate)
                    ) * $quantity;
                default:
                    throw new RuntimeException('Undefined tax calc method.');
            }
        }

        return $price * $quantity;
    }

    /**
     * 価格に対する消費税を計算する
     *
     * @param  int $price 価格
     * @param  double  $taxRate 税率
     *
     * @return int 消費税
     */
    public function calcTax($price, $taxRate)
    {
        switch ($this->Config->get('shopping-cart_tax-format', '')) {
            case 'extax':
                return $this->roundTax(
                    $this->calcTaxFromTaxExcluded($price, $taxRate)
                );
            case 'intax':
                return $this->roundTax(
                    $this->calcTaxFromTaxIncluded($price, $taxRate)
                );
            default:
                throw new RuntimeException('Undefined tax expression format.');
        }
    }
}
