import {
  ref, watch, computed, toRefs, reactive,
} from 'vue';
import {
  useAriaHidden, useBodyScrollLock, useFocusTrap, useMounted,
} from '../composables';

function isScrollable(el) {
  return el.scrollHeight > el.clientHeight || el.scrollWidth > el.clientWidth;
}

export default {
  name: 'Modal',
  template: '#Modal',
  props: {
    isOpen: {
      default: false,
      type: Boolean,
    },
    onClose: {
      default: () => {},
      type: Function,
    },
    container: {
      default: document.body,
      type: [String, HTMLElement],
    },
    preventScroll: {
      default: true,
      type: Boolean,
    },
    shouldCloseOnBackdropClick: {
      default: true,
      type: Boolean,
    },
    shouldCloseOnEsc: {
      default: true,
      type: Boolean,
    },
    closeTimeout: {
      default: 0,
      type: Number,
    },
    role: {
      default: 'dialog',
      type: String,
    },
    ariaModal: {
      default: true,
      type: Boolean,
    },
    ariaLabelledby: String,
    ariaDescribedby: String,
  },
  emits: ['afterOpen', 'afterClose'],
  setup(props, { emit }) {
    const dialogRef = ref(null);

    // Focus trap
    const { activate, deactivate } = useFocusTrap(dialogRef, {
      allowOutsideClick: true, // falseの場合、モーダルを閉じてもFocusTrapが解除されないので注意
      escapeDeactivates: false, // ESCキーを押したときの動作は自前でカスタマイズしたいのでfalse
      returnFocusOnDeactivate: true,
    });

    // Disable scroll of body
    const { disableScroll, enableScroll } = useBodyScrollLock(dialogRef);

    // Hides from ARIA everything, except modal dialog
    const { hide, unhide } = useAriaHidden(dialogRef);

    const isMounted = useMounted();
    const state = reactive({
      isOpen: false,
      beforeClose: false,
      afterOpen: false,
    });

    watch(
      toRefs(props).isOpen,
      (isOpen) => {
        if (isOpen) {
          state.isOpen = true;
        } else if (isMounted.value && state.isOpen) {
          state.beforeClose = true;
        }
      },
      { immediate: true },
    );

    watch(
      () => state.isOpen,
      (isOpen, _, onCleanup) => {
        let animationFrame;
        if (isOpen) {
          animationFrame = window.requestAnimationFrame(() => {
            state.afterOpen = true;
            if (isOpen) {
              emit('afterOpen');
            }
          });
        }

        onCleanup(() => {
          if (animationFrame) {
            cancelAnimationFrame(animationFrame);
          }
        });
      },
      { immediate: true },
    );

    watch(
      () => state.beforeClose,
      (beforeClose, _, onCleanup) => {
        let timeoutId;
        if (beforeClose) {
          timeoutId = window.setTimeout(() => {
            state.isOpen = false;
            state.beforeClose = false;
            state.afterOpen = false;
            if (props.preventScroll) {
              enableScroll();
            }
            deactivate();
            unhide();
            emit('afterClose');
          }, props.closeTimeout);
        }

        onCleanup(() => {
          if (timeoutId) {
            clearTimeout(timeoutId);
          }
        });
      },
      { immediate: true },
    );

    watch(
      () => state.afterOpen,
      (afterOpen) => {
        if (afterOpen) {
          if (props.preventScroll) {
            disableScroll({
              allowTouchMove(el) {
                // モーダルの内部はスクロールを許可する
                let element = el;
                while (element && element !== document.body) {
                  if (element !== null && isScrollable(element)) {
                    return true;
                  }

                  element = element.parentElement;
                }
                return false;
              },
            });
          }
          activate();
          hide();
        }
      },
      { immediate: true },
    );

    const shouldBeClosed = computed(() => !state.isOpen && !state.beforeClose);

    return {
      dialogRef,
      shouldBeClosed,
      state,
    };
  },
};
