import { ref, computed, reactive } from 'vue';
import { useStore } from 'vuex';

import { Modal, QuantitySelect } from '../../components';
import { isObject } from '../util';

export default {
  name: 'AddToCart',
  template: '#AddToCart',
  components: {
    Modal,
    QuantitySelect,
  },
  props: {
    eid: {
      type: Number,
      required: true,
    },
    title: {
      type: String,
      required: true,
    },
    stock: Number,
  },
  setup(props) {
    const added = ref(false);
    const setAdded = (value) => (added.value = value);

    const form = ref(null);

    const isProcessing = ref(false);
    const setIsProcessing = (value) => {
      isProcessing.value = value;
    };

    const store = useStore();

    const isLoading = computed(() => store.state.cart.isLoading);
    const item = computed(() => store.getters['cart/findItem'](props.eid));
    const remain = computed(() => {
      if (props.stock === null) {
        return Number.MAX_SAFE_INTEGER;
      }

      return item.value ? props.stock - item.value.quantity : props.stock;
    });

    // バリデーション
    const errors = reactive({
      eid: {
        required: false,
      },
      quantity: {
        required: false,
        integer: false,
        min: false,
        stock: false,
      },
    });

    const setError = (field, option, isValid) => {
      if (isObject(errors[field])) {
        errors[field][option] = isValid;
      }

      errors[field] = {
        [option]: isValid,
      };
    };

    const isError = computed(() => Object.values(errors).some((options) => Object.values(options).some((isValid) => isValid)));

    const resetErrors = () => {
      Object.entries(errors).forEach(([field, options]) => {
        Object.keys(options).forEach((option) => {
          setError(field, option, false);
        });
      });
    };

    const addToCart = async () => {
      try {
        resetErrors();
        setIsProcessing(true);
        await store.dispatch('cart/addToCart', new FormData(form.value));
        setAdded(true);
      } catch (e) {
        e.response.data.errors.forEach(({ field, option }) => {
          if (!field || !option) {
            return;
          }
          setError(field, option, true);
        });
      } finally {
        setIsProcessing(false);
      }
    };

    return {
      added,
      isLoading,
      isProcessing,
      form,
      isError,
      errors,
      item,
      remain,
      setAdded,
      resetErrors,
      addToCart,
    };
  },
};
