import {
  reactive, ref, onMounted,
} from 'vue';
import { useStore } from 'vuex';

import { formatMoney, getHeightFromRatio, isObject } from '../util';

export default {
  name: 'ShoppingCartItem',
  template: '#ShoppingCartItem',
  props: {
    item: {
      type: Object,
      required: true,
    },
    isEditing: {
      type: Boolean,
      default: false,
    },
    onEditButtonClick: {
      type: Function,
      default: () => {},
    },
    afterUpdate: {
      type: Function,
      default: () => {},
    },
    afterRemove: {
      type: Function,
      default: () => {},
    },
  },
  setup(props) {
    const root = ref(null);
    const updateForm = ref(null); // 数量更新フォーム
    const removeForm = ref(null); // 商品の削除フォーム
    const data = reactive({
      quantity: props.item.quantity,
    });

    const isProcessing = ref(false);
    const setIsProcessing = (value) => {
      isProcessing.value = value;
    };

    const store = useStore();

    onMounted(() => {
      if (ACMS !== undefined && typeof ACMS.Dispatch === 'function') {
        return ACMS.Dispatch(root.value);
      }

      return ACMS !== undefined && ACMS.Ready(() => {
        ACMS.Dispatch(root.value);
      });
    });

    // バリデーション
    const errors = reactive({
      eid: {
        required: false,
      },
      quantity: {
        required: false,
        integer: false,
        min: false,
        stock: false,
      },
    });

    const setError = (field, option, isValid) => {
      if (isObject(errors[field])) {
        return errors[field][option] = isValid;
      }

      return errors[field] = {
        [option]: isValid,
      };
    };

    const resetErrors = () => {
      Object.entries(errors).forEach(([field, options]) => {
        Object.keys(options).forEach((option) => {
          setError(field, option, false);
        });
      });
    };

    const validater = {
      quantity(field, value) {
        setError(field, 'required', !value);
        setError(field, 'integer', !/^\d+$/.test(value));
        setError(field, 'min', value <= 0);
        setError(field, 'stock', value > parseInt(props.item.item_stock, 10));
      },
    };

    const validate = (event) => {
      if (typeof validater[event.target.name] === 'function') {
        validater[event.target.name](event.target.name, event.target.value);
      }
    };

    const updateQuantity = async () => {
      try {
        resetErrors();
        setIsProcessing(true);
        await store.dispatch('cart/updateQuantity', new FormData(updateForm.value));
        props.afterUpdate();
      } catch (e) {
        e.response.data.errors.forEach(({ field, option }) => {
          if (!field || !option) {
            return;
          }
          setError(field, option, true);
        });
      } finally {
        setIsProcessing(false);
      }
    };

    const removeFromCart = async () => {
      try {
        resetErrors();
        setIsProcessing(true);
        await store.dispatch('cart/removeFromCart', new FormData(removeForm.value));
        props.afterRemove();
      } catch (e) {
        e.response.data.errors.forEach(({ field, option }) => {
          if (!field || !option) {
            return;
          }
          setError(field, option, true);
        });
      } finally {
        setIsProcessing(false);
      }
    };

    return {
      root,
      updateForm,
      removeForm,
      data,
      isProcessing,
      errors,
      formatMoney,
      getHeightFromRatio,
      validate,
      updateQuantity,
      removeFromCart,
    };
  },
};
