import {
  watch, ref, onUnmounted, unref,
} from 'vue';
import { createFocusTrap } from 'focus-trap';

/**
 * Reactive focus-trap
 *
 * @param targetRef The target element to trap focus within
 * @param options Focus trap options
 */
export const useFocusTrap = (targetRef, options = {}) => {
  let trap;

  const isActive = ref(false);
  const isPaused = ref(false);

  const activate = (opts) => trap && trap.activate(opts);
  const deactivate = (opts) => trap && trap.deactivate(opts);

  const pause = () => {
    if (trap) {
      trap.pause();
      isPaused.value = true;
    }
  };

  const unpause = () => {
    if (trap) {
      trap.unpause();
      isPaused.value = false;
    }
  };

  watch(
    () => unref(targetRef),
    (el) => {
      if (!el) {
        return;
      }

      trap = createFocusTrap(el, {
        ...options,
        onActivate() {
          isActive.value = true;

          if (options.onActivate) {
            options.onActivate();
          }
        },
        onDeactivate() {
          isActive.value = false;

          if (options.onDeactivate) {
            options.onDeactivate();
          }
        },
      });
    },
    { flush: 'post' },
  );

  onUnmounted(() => deactivate());

  return {
    isActive,
    isPaused,
    activate,
    deactivate,
    pause,
    unpause,
  };
};
